from collections.abc import Callable
from time import time

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
from bluesky.utils import MsgGenerator
from dodal.devices.eiger import EigerDetector
from dodal.devices.fast_grid_scan import FastGridScanCommon
from dodal.devices.synchrotron import Synchrotron
from dodal.devices.zocalo.zocalo_results import (
    ZOCALO_STAGE_GROUP,
)
from dodal.log import LOGGER
from dodal.plan_stubs.check_topup import check_topup_and_wait_if_necessary
from scanspec.core import AxesPoints, Axis

from mx_bluesky.common.parameters.constants import (
    PlanNameConstants,
)
from mx_bluesky.common.plans.read_hardware import read_hardware_for_zocalo
from mx_bluesky.common.utils.tracing import TRACER


def _wait_for_zocalo_to_stage_then_do_fgs(
    grid_scan_device: FastGridScanCommon,
    detector: EigerDetector,
    synchrotron: Synchrotron,
    during_collection_plan: Callable[[], MsgGenerator] | None = None,
):
    expected_images = yield from bps.rd(grid_scan_device.expected_images)
    exposure_sec_per_image = yield from bps.rd(detector.cam.acquire_time)  # type: ignore # Fix types in ophyd-async (https://github.com/DiamondLightSource/mx-bluesky/issues/855)
    LOGGER.info("waiting for topup if necessary...")
    yield from check_topup_and_wait_if_necessary(
        synchrotron,
        expected_images * exposure_sec_per_image,
        30.0,
    )

    # Make sure ZocaloResults queue is clear and ready to accept our new data. Zocalo MUST
    # have been staged using ZOCALO_STAGE_GROUP prior to this
    LOGGER.info("Waiting for Zocalo device queue to have been cleared...")
    yield from bps.wait(ZOCALO_STAGE_GROUP)

    # Triggers Zocalo if RE is subscribed to ZocaloCallback
    yield from read_hardware_for_zocalo(detector)
    LOGGER.info("Wait for all moves with no assigned group")
    yield from bps.wait()

    LOGGER.info("kicking off FGS")
    yield from bps.kickoff(grid_scan_device, wait=True)
    gridscan_start_time = time()
    if during_collection_plan:
        yield from during_collection_plan()
    LOGGER.info("completing FGS")
    yield from bps.complete(grid_scan_device, wait=True)
    # Remove this logging statement once metrics have been added
    LOGGER.info(
        f"Grid scan motion program took {round(time() - gridscan_start_time, 2)} to complete"
    )


def kickoff_and_complete_gridscan(
    gridscan: FastGridScanCommon,
    detector: EigerDetector,  # Once Eiger inherits from StandardDetector, use that type instead
    synchrotron: Synchrotron,
    scan_points: list[AxesPoints[Axis]],
    scan_start_indices: list[int],
    plan_during_collection: Callable[[], MsgGenerator] | None = None,
):
    """Triggers a grid scan motion program and waits for completion, accounting for synchrotron topup.
    If the RunEngine is subscribed to ZocaloCallback, this plan will also trigger Zocalo.

    Can be used for multiple successive grid scans, see Hyperion's usage

    Args:
        gridscan (FastGridScanCommon):          Device which can trigger a fast grid scan and wait for completion
        detector (EigerDetector)                Detector device
        synchrotron (Synchrotron):              Synchrotron device
        scan_points (list[AxesPoints[Axis]]):   Each element in the list contains all the grid points for that grid scan.
                                                Two elements in this list indicates that two grid scans will be done, eg for Hyperion's 3D grid scans.
        scan_start_indices (list[int]):         Contains the first index of each grid scan
        plan_during_collection (Optional, MsgGenerator): Generic plan called in between kickoff and completion,
                                                eg waiting on zocalo.
    """

    assert len(scan_points) == len(scan_start_indices), (
        "scan_points and scan_start_indices must be lists of the same length!"
    )

    plan_name = PlanNameConstants.DO_FGS

    @TRACER.start_as_current_span(plan_name)
    @bpp.set_run_key_decorator(plan_name)
    @bpp.run_decorator(
        md={
            "subplan_name": plan_name,
            "scan_points": scan_points,
            "scan_start_indices": scan_start_indices,
        }
    )
    @bpp.contingency_decorator(
        except_plan=lambda e: (yield from bps.stop(detector)),  # type: ignore # Fix types in ophyd-async (https://github.com/DiamondLightSource/mx-bluesky/issues/855)
        else_plan=lambda: (yield from bps.unstage(detector)),
    )
    def _decorated_do_fgs():
        yield from _wait_for_zocalo_to_stage_then_do_fgs(
            gridscan,
            detector,
            synchrotron,
            during_collection_plan=plan_during_collection,
        )

    yield from _decorated_do_fgs()



from unittest.mock import MagicMock

import pytest
from bluesky import plan_stubs as bps
from bluesky.utils import FailedStatus
from dodal.beamlines import i03
from ophyd.status import Status
from ophyd_async.testing import get_mock_put

from mx_bluesky.hyperion.device_setup_plans.utils import (
    start_preparing_data_collection_then_do_plan,
)


@pytest.fixture()
def mock_eiger():
    eiger = i03.eiger(mock=True)
    eiger.detector_params = MagicMock()
    eiger.async_stage = MagicMock()
    eiger.disarm_detector = MagicMock()
    return eiger


class MyTestException(Exception):
    pass


def test_given_plan_raises_when_exception_raised_then_eiger_disarmed_and_correct_exception_returned(
    beamstop_i03, mock_eiger, detector_motion, RE
):
    def my_plan():
        yield from bps.null()
        raise MyTestException()

    eiger = mock_eiger

    with pytest.raises(MyTestException):
        RE(
            start_preparing_data_collection_then_do_plan(
                beamstop_i03, eiger, detector_motion, 100, my_plan()
            )
        )

    # Check detector was armed
    eiger.async_stage.assert_called_once()

    eiger.disarm_detector.assert_called_once()


@pytest.fixture()
def null_plan():
    yield from bps.null()


def test_given_shutter_open_fails_then_eiger_disarmed_and_correct_exception_returned(
    beamstop_i03, mock_eiger, null_plan, RE
):
    detector_motion = MagicMock()
    status = Status()
    status.set_exception(MyTestException())
    detector_motion.z.set = MagicMock(return_value=status)

    with pytest.raises(FailedStatus) as e:
        RE(
            start_preparing_data_collection_then_do_plan(
                beamstop_i03, mock_eiger, detector_motion, 100, null_plan
            )
        )
    assert e.value.args[0] is status

    mock_eiger.async_stage.assert_called_once()
    detector_motion.z.set.assert_called_once()
    mock_eiger.disarm_detector.assert_called_once()


def test_given_detector_move_fails_then_eiger_disarmed_and_correct_exception_returned(
    beamstop_i03, mock_eiger, detector_motion, null_plan, RE
):
    status = Status()
    status.set_exception(MyTestException())
    detector_motion.shutter.set = MagicMock(return_value=status)

    with pytest.raises(FailedStatus) as e:
        RE(
            start_preparing_data_collection_then_do_plan(
                beamstop_i03, mock_eiger, detector_motion, 100, null_plan
            )
        )
    assert e.value.args[0] is status
    mock_eiger.async_stage.assert_called_once()

    def wait_for_set():
        yield from bps.wait(group="ready_for_data_collection")

    RE(wait_for_set())
    get_mock_put(detector_motion.z.user_setpoint).assert_called_once()
    mock_eiger.disarm_detector.assert_called_once()



import numpy as np
from bluesky import plan_stubs as bps
from dodal.devices.smargon import Smargon

from mx_bluesky.common.utils.exceptions import SampleException


def move_smargon_warn_on_out_of_range(
    smargon: Smargon, position: np.ndarray | list[float] | tuple[float, float, float]
):
    """Throws a SampleException if the specified position is out of range for the
    smargon. Otherwise moves to that position."""
    limits = yield from smargon.get_xyz_limits()
    if not limits.position_valid(position):
        raise SampleException(
            "Pin tip centring failed - pin too long/short/bent and out of range"
        )
    yield from bps.mv(
        smargon.x,
        position[0],
        smargon.y,
        position[1],
        smargon.z,
        position[2],
    )



from functools import partial
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from bluesky.utils import Msg
from dodal.devices.aperturescatterguard import ApertureScatterguard, ApertureValue
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.smargon import Smargon, StubPosition
from dodal.devices.webcam import Webcam
from ophyd.sim import NullStatus
from ophyd_async.testing import set_mock_value

from mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy import (
    RobotLoadAndEnergyChangeComposite,
    prepare_for_robot_load,
    robot_load_and_change_energy_plan,
    take_robot_snapshots,
)
from mx_bluesky.hyperion.external_interaction.callbacks.robot_load.ispyb_callback import (
    RobotLoadISPyBCallback,
)
from mx_bluesky.hyperion.parameters.robot_load import RobotLoadAndEnergyChange

from ....conftest import raw_params_from_file


@pytest.fixture
def robot_load_and_energy_change_params():
    params = raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_robot_load_params.json"
    )
    return RobotLoadAndEnergyChange(**params)


@pytest.fixture
def robot_load_and_energy_change_params_no_energy(robot_load_and_energy_change_params):
    robot_load_and_energy_change_params.demand_energy_ev = None
    return robot_load_and_energy_change_params


def dummy_set_energy_plan(energy, composite):
    return (yield Msg("set_energy_plan"))


@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy.set_energy_plan",
    MagicMock(side_effect=dummy_set_energy_plan),
)
def test_when_plan_run_with_requested_energy_specified_energy_change_executes(
    robot_load_and_energy_change_composite: RobotLoadAndEnergyChangeComposite,
    robot_load_and_energy_change_params: RobotLoadAndEnergyChange,
    sim_run_engine: RunEngineSimulator,
):
    sim_run_engine.add_handler(
        "read",
        lambda msg: {"dcm-energy_in_kev": {"value": 11.105}},
        "dcm-energy_in_kev",
    )
    messages = sim_run_engine.simulate_plan(
        robot_load_and_change_energy_plan(
            robot_load_and_energy_change_composite, robot_load_and_energy_change_params
        )
    )
    assert_message_and_return_remaining(
        messages, lambda msg: msg.command == "set_energy_plan"
    )


def run_simulating_smargon_wait(
    robot_load_then_centre_params,
    robot_load_composite,
    total_disabled_reads,
    sim_run_engine: RunEngineSimulator,
):
    num_of_reads = 0

    def return_not_disabled_after_reads(_):
        nonlocal num_of_reads
        num_of_reads += 1
        return {"values": {"value": int(num_of_reads < total_disabled_reads)}}

    sim_run_engine.add_handler(
        "locate",
        lambda msg: {"readback": 11.105},
        "dcm-energy_in_kev",
    )
    sim_run_engine.add_handler(
        "read", return_not_disabled_after_reads, "smargon-disabled"
    )

    return sim_run_engine.simulate_plan(
        robot_load_and_change_energy_plan(
            robot_load_composite, robot_load_then_centre_params
        )
    )


@pytest.mark.parametrize("total_disabled_reads", [5, 3, 14])
@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy.set_energy_plan",
    MagicMock(return_value=iter([])),
)
def test_given_smargon_disabled_when_plan_run_then_waits_on_smargon(
    robot_load_and_energy_change_composite: RobotLoadAndEnergyChangeComposite,
    robot_load_and_energy_change_params: RobotLoadAndEnergyChange,
    total_disabled_reads: int,
    sim_run_engine,
):
    messages = run_simulating_smargon_wait(
        robot_load_and_energy_change_params,
        robot_load_and_energy_change_composite,
        total_disabled_reads,
        sim_run_engine,
    )

    sleep_messages = filter(lambda msg: msg.command == "sleep", messages)
    read_disabled_messages = filter(
        lambda msg: msg.command == "read" and msg.obj.name == "smargon-disabled",
        messages,
    )

    assert len(list(sleep_messages)) == total_disabled_reads - 1
    assert len(list(read_disabled_messages)) == total_disabled_reads


@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy.set_energy_plan",
    MagicMock(return_value=iter([])),
)
def test_given_smargon_disabled_for_longer_than_timeout_when_plan_run_then_throws_exception(
    robot_load_and_energy_change_composite: RobotLoadAndEnergyChangeComposite,
    robot_load_and_energy_change_params: RobotLoadAndEnergyChange,
    sim_run_engine,
):
    with pytest.raises(TimeoutError):
        run_simulating_smargon_wait(
            robot_load_and_energy_change_params,
            robot_load_and_energy_change_composite,
            1000,
            sim_run_engine,
        )


async def test_when_prepare_for_robot_load_called_then_moves_as_expected(
    aperture_scatterguard: ApertureScatterguard, smargon: Smargon, done_status
):
    smargon.stub_offsets.set = MagicMock(return_value=done_status)
    aperture_scatterguard.set = MagicMock(return_value=done_status)

    set_mock_value(smargon.x.user_setpoint, 10)
    set_mock_value(smargon.z.user_setpoint, 5)
    set_mock_value(smargon.omega.user_setpoint, 90)

    RE = RunEngine()
    RE(prepare_for_robot_load(aperture_scatterguard, smargon))

    assert await smargon.x.user_setpoint.get_value() == 0
    assert await smargon.z.user_setpoint.get_value() == 0
    assert await smargon.omega.user_setpoint.get_value() == 0

    smargon.stub_offsets.set.assert_called_once_with(StubPosition.RESET_TO_ROBOT_LOAD)  # type: ignore
    aperture_scatterguard.set.assert_called_once_with(ApertureValue.OUT_OF_BEAM)  # type: ignore


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.robot_load.ispyb_callback.ExpeyeInteraction"
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy.set_energy_plan",
    MagicMock(return_value=iter([])),
)
def test_given_ispyb_callback_attached_when_robot_load_then_centre_plan_called_then_ispyb_deposited(
    exp_eye: MagicMock,
    robot_load_and_energy_change_composite: RobotLoadAndEnergyChangeComposite,
    robot_load_and_energy_change_params: RobotLoadAndEnergyChange,
):
    set_mock_value(
        robot_load_and_energy_change_composite.oav.snapshot.last_saved_path,
        "test_oav_snapshot",
    )  # type: ignore
    set_mock_value(
        robot_load_and_energy_change_composite.webcam.last_saved_path,
        "test_webcam_snapshot",
    )
    robot_load_and_energy_change_composite.webcam.trigger = MagicMock(
        return_value=NullStatus()
    )

    RE = RunEngine()
    RE.subscribe(RobotLoadISPyBCallback())

    action_id = 1098
    exp_eye.return_value.start_load.return_value = action_id

    RE(
        robot_load_and_change_energy_plan(
            robot_load_and_energy_change_composite, robot_load_and_energy_change_params
        )
    )

    exp_eye.return_value.start_load.assert_called_once_with("cm31105", 4, 12345, 40, 3)
    exp_eye.return_value.update_barcode_and_snapshots.assert_called_once_with(
        action_id, "BARCODE", "test_webcam_snapshot", "test_oav_snapshot"
    )
    exp_eye.return_value.end_load.assert_called_once_with(action_id, "success", "OK")


@patch("mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy.datetime")
async def test_when_take_snapshots_called_then_filename_and_directory_set_and_device_triggered(
    mock_datetime: MagicMock, oav: OAV, webcam: Webcam
):
    TEST_DIRECTORY = "TEST"

    mock_datetime.now.return_value.strftime.return_value = "TIME"

    RE = RunEngine()
    oav.snapshot.trigger = MagicMock(side_effect=oav.snapshot.trigger)
    webcam.trigger = MagicMock(return_value=NullStatus())

    RE(take_robot_snapshots(oav, webcam, Path(TEST_DIRECTORY)))

    oav.snapshot.trigger.assert_called_once()
    assert await oav.snapshot.filename.get_value() == "TIME_oav-snapshot_after_load"
    assert await oav.snapshot.directory.get_value() == TEST_DIRECTORY

    webcam.trigger.assert_called_once()
    assert (await webcam.filename.get_value()) == "TIME_webcam_after_load"
    assert (await webcam.directory.get_value()) == TEST_DIRECTORY


def test_given_lower_gonio_moved_when_robot_load_then_lower_gonio_moved_to_home_and_back(
    robot_load_and_energy_change_composite: RobotLoadAndEnergyChangeComposite,
    robot_load_and_energy_change_params_no_energy: RobotLoadAndEnergyChange,
    sim_run_engine: RunEngineSimulator,
):
    initial_values = {"x": 0.11, "y": 0.12, "z": 0.13}

    def get_read(axis, msg):
        return {"readback": initial_values[axis]}

    for axis in initial_values.keys():
        sim_run_engine.add_handler(
            "locate", partial(get_read, axis), f"lower_gonio-{axis}"
        )

    messages = sim_run_engine.simulate_plan(
        robot_load_and_change_energy_plan(
            robot_load_and_energy_change_composite,
            robot_load_and_energy_change_params_no_energy,
        )
    )

    for axis in initial_values.keys():
        messages = assert_message_and_return_remaining(
            messages,
            lambda msg: msg.command == "set"
            and msg.obj.name == f"lower_gonio-{axis}"
            and msg.args == (0,),
        )

    for axis, initial in initial_values.items():
        messages = assert_message_and_return_remaining(
            messages,
            lambda msg: msg.command == "set"
            and msg.obj.name == f"lower_gonio-{axis}"
            and msg.args == (initial,),
        )


@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy.set_energy_plan",
    MagicMock(return_value=iter([])),
)
def test_when_plan_run_then_lower_gonio_moved_before_robot_loads_and_back_after_smargon_enabled(
    robot_load_and_energy_change_composite: RobotLoadAndEnergyChangeComposite,
    robot_load_and_energy_change_params_no_energy: RobotLoadAndEnergyChange,
    sim_run_engine: RunEngineSimulator,
):
    initial_values = {"x": 0.11, "y": 0.12, "z": 0.13}

    def get_read(axis, msg):
        return {"readback": initial_values[axis]}

    for axis in initial_values.keys():
        sim_run_engine.add_handler(
            "locate", partial(get_read, axis), f"lower_gonio-{axis}"
        )

    messages = sim_run_engine.simulate_plan(
        robot_load_and_change_energy_plan(
            robot_load_and_energy_change_composite,
            robot_load_and_energy_change_params_no_energy,
        )
    )

    assert_message_and_return_remaining(
        messages, lambda msg: msg.command == "set" and msg.obj.name == "robot"
    )

    for axis in initial_values.keys():
        messages = assert_message_and_return_remaining(
            messages,
            lambda msg: msg.command == "set"
            and msg.obj.name == f"lower_gonio-{axis}"
            and msg.args == (0,),
        )

    assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "read" and msg.obj.name == "smargon-disabled",
    )

    for axis, initial in initial_values.items():
        messages = assert_message_and_return_remaining(
            messages,
            lambda msg: msg.command == "set"
            and msg.obj.name == f"lower_gonio-{axis}"  # noqa
            and msg.args == (initial,),  # noqa
        )


@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy.set_energy_plan",
    MagicMock(return_value=iter([])),
)
def test_when_plan_run_then_thawing_turned_on_for_expected_time(
    robot_load_and_energy_change_composite: RobotLoadAndEnergyChangeComposite,
    robot_load_and_energy_change_params_no_energy: RobotLoadAndEnergyChange,
    sim_run_engine: RunEngineSimulator,
):
    robot_load_and_energy_change_params_no_energy.thawing_time = (thaw_time := 50)

    sim_run_engine.add_handler(
        "read",
        lambda msg: {"dcm-energy_in_kev": {"value": 11.105}},
        "dcm-energy_in_kev",
    )

    messages = sim_run_engine.simulate_plan(
        robot_load_and_change_energy_plan(
            robot_load_and_energy_change_composite,
            robot_load_and_energy_change_params_no_energy,
        )
    )

    assert_message_and_return_remaining(
        messages,
        lambda msg: msg.command == "set"
        and msg.obj.name == "thawer-thaw_for_time_s"
        and msg.args[0] == thaw_time,
    )



from datetime import datetime
from typing import Protocol

from bluesky import plan_stubs as bps
from bluesky.utils import MsgGenerator
from dodal.devices.aperturescatterguard import ApertureScatterguard, ApertureValue
from dodal.devices.backlight import Backlight, BacklightPosition
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.oav_parameters import OAVParameters
from dodal.devices.smargon import Smargon

from mx_bluesky.common.parameters.components import WithSnapshot
from mx_bluesky.hyperion.device_setup_plans.setup_oav import setup_general_oav_params
from mx_bluesky.hyperion.parameters.constants import CONST, DocDescriptorNames

OAV_SNAPSHOT_SETUP_SHOT = "oav_snapshot_setup_shot"
OAV_SNAPSHOT_GROUP = "oav_snapshot_group"


class OavSnapshotComposite(Protocol):
    smargon: Smargon
    oav: OAV
    aperture_scatterguard: ApertureScatterguard
    backlight: Backlight


def setup_beamline_for_OAV(
    smargon: Smargon,
    backlight: Backlight,
    aperture_scatterguard: ApertureScatterguard,
    group=CONST.WAIT.READY_FOR_OAV,
):
    max_vel = yield from bps.rd(smargon.omega.max_velocity)
    yield from bps.abs_set(smargon.omega.velocity, max_vel, group=group)
    yield from bps.abs_set(backlight, BacklightPosition.IN, group=group)
    yield from bps.abs_set(
        aperture_scatterguard, ApertureValue.OUT_OF_BEAM, group=group
    )


def oav_snapshot_plan(
    composite: OavSnapshotComposite,
    parameters: WithSnapshot,
    oav_parameters: OAVParameters,
) -> MsgGenerator:
    if not parameters.take_snapshots:
        return
    yield from _setup_oav(composite, parameters, oav_parameters)
    for omega in parameters.snapshot_omegas_deg or []:
        yield from _take_oav_snapshot(composite, omega)


def _setup_oav(
    composite: OavSnapshotComposite,
    parameters: WithSnapshot,
    oav_parameters: OAVParameters,
):
    yield from setup_general_oav_params(composite.oav, oav_parameters)
    yield from bps.abs_set(
        composite.oav.snapshot.directory,
        str(parameters.snapshot_directory),
    )


def _take_oav_snapshot(composite: OavSnapshotComposite, omega: float):
    yield from bps.abs_set(
        composite.smargon.omega, omega, group=OAV_SNAPSHOT_SETUP_SHOT
    )
    time_now = datetime.now()
    filename = f"{time_now.strftime('%H%M%S%f')[:8]}_oav_snapshot_{omega:.0f}"
    yield from bps.abs_set(
        composite.oav.snapshot.filename,
        filename,
        group=OAV_SNAPSHOT_SETUP_SHOT,
    )
    yield from bps.wait(group=OAV_SNAPSHOT_SETUP_SHOT)
    yield from bps.trigger(composite.oav.snapshot, wait=True)
    yield from bps.create(DocDescriptorNames.OAV_ROTATION_SNAPSHOT_TRIGGERED)
    yield from bps.read(composite.oav)
    yield from bps.save()



from unittest.mock import MagicMock, call, patch

import pytest
from dodal.devices.zocalo import ZocaloStartInfo

from mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback import (
    ZocaloCallback,
    ZocaloTrigger,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import (
    IspybIds,
    StoreInIspyb,
)
from mx_bluesky.common.utils.exceptions import ISPyBDepositionNotMade
from mx_bluesky.hyperion.external_interaction.callbacks.__main__ import (
    create_gridscan_callbacks,
)

from .....conftest import TestData

EXPECTED_RUN_START_MESSAGE = {"subplan_name": "test_plan_name", "uid": "my_uuid"}
EXPECTED_RUN_END_MESSAGE = {"event": "end", "run_start": "my_uuid"}

td = TestData()


class TestZocaloHandler:
    def _setup_handler(self):
        zocalo_handler = ZocaloCallback("test_plan_name", "test_env")
        return zocalo_handler

    def test_handler_doesnt_trigger_on_wrong_plan(self):
        zocalo_handler = self._setup_handler()
        zocalo_handler.start({"sybplan_name": "_not_test_plan_name"})  # type: ignore
        assert len(zocalo_handler.zocalo_info) == 0

    @patch(
        "mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback.ZocaloTrigger",
        autospec=True,
    )
    def test_handler_stores_collection_if_ispyb_ids_come_in_with_triggering_plan(
        self, zocalo_trigger: ZocaloTrigger
    ):
        zocalo_handler = self._setup_handler()
        assert not zocalo_handler.zocalo_info
        zocalo_handler.start(
            {
                **EXPECTED_RUN_START_MESSAGE,
                "ispyb_dcids": (135, 139),
                "scan_points": [{"test": [1, 2, 3]}, {"test": [2, 3, 4]}],
            }  # type: ignore
        )
        assert len(zocalo_handler.zocalo_info) == 2

    @patch(
        "mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback.ZocaloTrigger",
        autospec=True,
    )
    def test_handler_stores_collection_ispyb_ids_come_in_as_subplan(
        self, zocalo_trigger: ZocaloTrigger
    ):
        zocalo_handler = self._setup_handler()
        assert not zocalo_handler.zocalo_info
        zocalo_handler.start(EXPECTED_RUN_START_MESSAGE)  # type: ignore
        assert not zocalo_handler.zocalo_info
        zocalo_handler.start(
            {
                "subplan_name": "other_plan",
                "ispyb_dcids": (135,),
                "scan_points": [{"test": [1, 2, 3]}],
            }  # type: ignore
        )
        zocalo_handler.start(
            {
                "subplan_name": "other_plan",
                "ispyb_dcids": (563,),
                "scan_points": [{"test": [2, 3, 4]}],
            }  # type: ignore
        )

        assert len(zocalo_handler.zocalo_info) == 2

    @patch(
        "mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback.ZocaloTrigger",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.external_interaction.callbacks.xray_centre.nexus_callback.NexusWriter",
    )
    @patch(
        "mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback.StoreInIspyb",
    )
    def test_execution_of_do_fgs_triggers_zocalo_calls(
        self, ispyb_store: MagicMock, nexus_writer: MagicMock, zocalo_trigger
    ):
        dc_ids = (1, 2)
        dcg_id = 4

        mock_ids = IspybIds(data_collection_ids=dc_ids, data_collection_group_id=dcg_id)
        ispyb_store.return_value.mock_add_spec(StoreInIspyb)

        _, ispyb_cb = create_gridscan_callbacks()
        ispyb_cb.active = True
        assert isinstance(zocalo_handler := ispyb_cb.emit_cb, ZocaloCallback)
        zocalo_handler._reset_state()
        zocalo_handler._reset_state = MagicMock()

        ispyb_store.return_value.begin_deposition.return_value = mock_ids
        ispyb_store.return_value.update_deposition.return_value = mock_ids

        ispyb_cb.start(td.test_gridscan3d_start_document)  # type: ignore
        ispyb_cb.start(td.test_gridscan_outer_start_document)  # type: ignore
        ispyb_cb.start(td.test_do_fgs_start_document)  # type: ignore
        ispyb_cb.descriptor(td.test_descriptor_document_pre_data_collection)  # type: ignore
        ispyb_cb.event(td.test_event_document_pre_data_collection)
        ispyb_cb.descriptor(td.test_descriptor_document_zocalo_hardware)
        ispyb_cb.event(td.test_event_document_zocalo_hardware)
        ispyb_cb.descriptor(
            td.test_descriptor_document_during_data_collection  # type: ignore
        )
        ispyb_cb.event(td.test_event_document_during_data_collection)
        assert zocalo_handler.zocalo_interactor is not None

        expected_start_calls = [
            call(ZocaloStartInfo(1, "test_path", 0, 200, 0)),
            call(ZocaloStartInfo(2, "test_path", 200, 300, 1)),
        ]

        zocalo_handler.zocalo_interactor.run_start.assert_has_calls(  # type: ignore
            expected_start_calls
        )
        assert zocalo_handler.zocalo_interactor.run_start.call_count == len(dc_ids)  # type: ignore

        ispyb_cb.stop(td.test_stop_document)

        zocalo_handler.zocalo_interactor.run_end.assert_has_calls(  # type: ignore
            [call(x) for x in dc_ids]
        )
        assert zocalo_handler.zocalo_interactor.run_end.call_count == len(dc_ids)  # type: ignore

        zocalo_handler._reset_state.assert_called()

    @patch(
        "mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback.ZocaloTrigger",
        autospec=True,
    )
    def test_handler_raises_on_the_end_of_a_plan_with_no_depositions(
        self, zocalo_trigger: ZocaloTrigger
    ):
        zocalo_handler = self._setup_handler()
        zocalo_handler.start(EXPECTED_RUN_START_MESSAGE)  # type: ignore
        with pytest.raises(ISPyBDepositionNotMade):
            zocalo_handler.stop(EXPECTED_RUN_END_MESSAGE)  # type: ignore



from functools import partial

import bluesky.plan_stubs as bps
from dodal.devices.areadetector.plugins.CAM import ColorMode
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.oav_parameters import OAVParameters
from dodal.devices.oav.pin_image_recognition import PinTipDetection

from mx_bluesky.hyperion.parameters.constants import CONST

# Helper function to make sure we set the waiting groups correctly
set_using_group = partial(bps.abs_set, group=CONST.WAIT.READY_FOR_OAV)


def setup_pin_tip_detection_params(
    pin_tip_detect_device: PinTipDetection, parameters: OAVParameters
):
    # select which blur to apply to image
    yield from set_using_group(
        pin_tip_detect_device.preprocess_operation, parameters.preprocess
    )

    # sets length scale for blurring
    yield from set_using_group(
        pin_tip_detect_device.preprocess_ksize, parameters.preprocess_K_size
    )

    # Canny edge detect - lower
    yield from set_using_group(
        pin_tip_detect_device.canny_lower_threshold,
        parameters.canny_edge_lower_threshold,
    )

    # Canny edge detect - upper
    yield from set_using_group(
        pin_tip_detect_device.canny_upper_threshold,
        parameters.canny_edge_upper_threshold,
    )

    # "Close" morphological operation
    yield from set_using_group(
        pin_tip_detect_device.close_ksize, parameters.close_ksize
    )

    # Sample detection direction
    yield from set_using_group(
        pin_tip_detect_device.scan_direction, parameters.direction
    )

    # Minimum height
    yield from set_using_group(
        pin_tip_detect_device.min_tip_height,
        parameters.minimum_height,
    )


def setup_general_oav_params(oav: OAV, parameters: OAVParameters):
    yield from set_using_group(oav.cam.color_mode, ColorMode.RGB1)
    yield from set_using_group(oav.cam.acquire_period, parameters.acquire_period)
    yield from set_using_group(oav.cam.acquire_time, parameters.exposure)
    yield from set_using_group(oav.cam.gain, parameters.gain)

    zoom_level_str = f"{float(parameters.zoom)}x"
    yield from bps.abs_set(
        oav.zoom_controller,
        zoom_level_str,
        wait=True,
    )


def pre_centring_setup_oav(
    oav: OAV,
    parameters: OAVParameters,
    pin_tip_detection_device: PinTipDetection,
):
    """
    Setup OAV PVs with required values.
    """
    yield from setup_general_oav_params(oav, parameters)
    yield from setup_pin_tip_detection_params(pin_tip_detection_device, parameters)
    yield from bps.wait(CONST.WAIT.READY_FOR_OAV)



from unittest.mock import DEFAULT, MagicMock, patch

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
import pytest
from bluesky.run_engine import RunEngine
from event_model.documents import Event, EventDescriptor, RunStart, RunStop

from mx_bluesky.common.external_interaction.callbacks.common.plan_reactive_callback import (
    PlanReactiveCallback,
)

from ..conftest import (
    MockReactiveCallback,
    get_test_plan,
)


def test_activity_gated_functions_not_called_when_inactive(
    mocked_test_callback: MockReactiveCallback,
):
    mocked_test_callback.start({})  # type: ignore
    mocked_test_callback.activity_gated_start.assert_not_called()  # type: ignore
    mocked_test_callback.descriptor({})  # type: ignore
    mocked_test_callback.activity_gated_descriptor.assert_not_called()  # type: ignore
    mocked_test_callback.event({})  # type: ignore
    mocked_test_callback.activity_gated_event.assert_not_called()  # type: ignore
    mocked_test_callback.stop({})  # type: ignore
    mocked_test_callback.activity_gated_stop.assert_not_called()  # type: ignore


def test_activity_gated_functions_called_when_active(
    mocked_test_callback: MockReactiveCallback,
):
    mocked_test_callback.active = True
    mocked_test_callback.start({})  # type: ignore
    mocked_test_callback.activity_gated_start.assert_called_once()  # type: ignore
    mocked_test_callback.descriptor({})  # type: ignore
    mocked_test_callback.activity_gated_descriptor.assert_called_once()  # type: ignore
    mocked_test_callback.event({})  # type: ignore
    mocked_test_callback.activity_gated_event.assert_called_once()  # type: ignore
    mocked_test_callback.stop({})  # type: ignore
    mocked_test_callback.activity_gated_stop.assert_called_once()  # type: ignore


def test_activates_on_appropriate_start_doc(mocked_test_callback):
    assert mocked_test_callback.active is False
    mocked_test_callback.start({"activate_callbacks": ["MockReactiveCallback"]})
    assert mocked_test_callback.active is True


def test_deactivates_on_appropriate_stop_doc_uid(mocked_test_callback):
    assert mocked_test_callback.active is False
    mocked_test_callback.start(
        {"activate_callbacks": ["MockReactiveCallback"], "uid": "foo"}
    )
    assert mocked_test_callback.active is True
    mocked_test_callback.stop({"run_start": "foo"})
    assert mocked_test_callback.active is False


def test_doesnt_deactivate_on_inappropriate_stop_doc_uid(mocked_test_callback):
    assert mocked_test_callback.active is False
    mocked_test_callback.start(
        {"activate_callbacks": ["MockReactiveCallback"], "uid": "foo"}
    )
    assert mocked_test_callback.active is True
    mocked_test_callback.stop({"run_start": "bar"})
    assert mocked_test_callback.active is True


def test_activates_on_metadata(
    RE_with_mock_callback: tuple[RunEngine, MockReactiveCallback],
):
    RE, callback = RE_with_mock_callback
    RE(get_test_plan("MockReactiveCallback")[0]())
    callback.activity_gated_start.assert_called_once()
    callback.activity_gated_descriptor.assert_called_once()
    callback.activity_gated_event.assert_called_once()
    callback.activity_gated_stop.assert_called_once()


def test_deactivates_after_closing(
    RE_with_mock_callback: tuple[RunEngine, MockReactiveCallback],
):
    RE, callback = RE_with_mock_callback
    assert callback.active is False
    RE(get_test_plan("MockReactiveCallback")[0]())
    assert callback.active is False


def test_doesnt_activate_on_wrong_metadata(
    RE_with_mock_callback: tuple[RunEngine, MockReactiveCallback],
):
    RE, callback = RE_with_mock_callback
    RE(get_test_plan("TestNotCallback")[0]())
    callback.activity_gated_start.assert_not_called()  # type: ignore
    callback.activity_gated_descriptor.assert_not_called()  # type: ignore
    callback.activity_gated_event.assert_not_called()  # type: ignore
    callback.activity_gated_stop.assert_not_called()  # type: ignore


def test_cb_logs_and_raises_exception():
    cb = MockReactiveCallback()
    cb.active = True

    class MockTestException(Exception): ...

    e = MockTestException()

    def mock_excepting_func(_):
        raise e

    cb.log = MagicMock()

    with pytest.raises(MockTestException):
        cb._run_activity_gated("start", mock_excepting_func, {"start": "test"})

    cb.log.exception.assert_called_with(e)


def test_emit_called_correctly():
    receiving_cb = MockReactiveCallback()
    test_cb = PlanReactiveCallback(emit=receiving_cb, log=MagicMock())

    start_doc: RunStart = {"uid": "123", "time": 0}
    desc_doc: EventDescriptor = {
        "data_keys": {},
        "run_start": "123",
        "uid": "987",
        "time": 0,
    }
    event_doc: Event = {
        "data": {},
        "time": 0,
        "descriptor": "987",
        "timestamps": {},
        "uid": "999",
        "seq_num": 0,
    }
    stop_doc: RunStop = {
        "exit_status": "success",
        "run_start": "123",
        "uid": "456",
        "time": 0,
    }

    test_cb.active = True
    receiving_cb.active = True

    test_cb.start(start_doc)
    receiving_cb.activity_gated_start.assert_called_once_with(start_doc)
    test_cb.descriptor(desc_doc)
    receiving_cb.activity_gated_descriptor.assert_called_once_with(desc_doc)
    test_cb.event(event_doc)
    receiving_cb.activity_gated_event.assert_called_once_with(event_doc)
    test_cb.stop(stop_doc)
    receiving_cb.activity_gated_stop.assert_called_once_with(stop_doc)


class OuterCallback(PlanReactiveCallback):
    pass


class InnerCallback(PlanReactiveCallback):
    pass


def test_activate_callbacks_doesnt_deactivate_unlisted_callbacks(RE: RunEngine):
    @bpp.set_run_key_decorator("inner_plan")
    @bpp.run_decorator(md={"activate_callbacks": ["InnerCallback"]})
    def inner_plan():
        yield from bps.null()

    @bpp.set_run_key_decorator("outer_plan")
    @bpp.run_decorator(md={"activate_callbacks": ["OuterCallback"]})
    def outer_plan():
        yield from inner_plan()

    outer_callback = OuterCallback(MagicMock())
    inner_callback = InnerCallback(MagicMock())

    RE.subscribe(outer_callback)
    RE.subscribe(inner_callback)

    with patch.multiple(
        outer_callback, activity_gated_start=DEFAULT, activity_gated_stop=DEFAULT
    ):
        with patch.multiple(
            inner_callback, activity_gated_start=DEFAULT, activity_gated_stop=DEFAULT
        ):
            root_mock = MagicMock()
            # fmt: off
            root_mock.attach_mock(outer_callback.activity_gated_start, "outer_start")  # pyright: ignore
            root_mock.attach_mock(outer_callback.activity_gated_stop, "outer_stop")  # pyright: ignore
            root_mock.attach_mock(inner_callback.activity_gated_start, "inner_start")  # pyright: ignore
            root_mock.attach_mock(inner_callback.activity_gated_stop, "inner_stop")  # pyright: ignore
            # fmt: on
            RE(outer_plan())

            assert [call[0] for call in root_mock.mock_calls] == [
                "outer_start",
                "outer_start",
                "inner_start",
                "outer_stop",
                "inner_stop",
                "outer_stop",
            ]

            assert (
                root_mock.mock_calls[0].args[0]["uid"]
                != root_mock.mock_calls[1].args[0]["uid"]
            )
            assert root_mock.mock_calls[1].args[0] == root_mock.mock_calls[2].args[0]
            assert root_mock.mock_calls[3].args[0] == root_mock.mock_calls[4].args[0]
            assert (
                root_mock.mock_calls[0].args[0]["uid"]
                == root_mock.mock_calls[5].args[0]["run_start"]
            )
            assert (
                root_mock.mock_calls[2].args[0]["uid"]
                == root_mock.mock_calls[4].args[0]["run_start"]
            )



from bluesky import preprocessors as bpp
from bluesky.preprocessors import plan_mutator
from bluesky.utils import Msg, MsgGenerator, make_decorator

from mx_bluesky.common.device_setup_plans.xbpm_feedback import (
    check_and_pause_feedback,
    unpause_xbpm_feedback_and_set_transmission_to_1,
)
from mx_bluesky.common.parameters.constants import PlanNameConstants
from mx_bluesky.common.protocols.protocols import (
    XBPMPauseDevices,
)


def transmission_and_xbpm_feedback_for_collection_wrapper(
    plan: MsgGenerator,
    devices: XBPMPauseDevices,
    desired_transmission_fraction: float,
    run_key_to_wrap: PlanNameConstants | None = None,
):
    """
    Sets the transmission for the data collection, ensuring the xbpm feedback is valid, then resets it immediately after
    the run has finished.

    This wrapper should be attached to the entry point of any beamline-specific plan that may disrupt the XBPM feedback,
    such as a data collection or an x-ray center grid scan.
    This wrapper will do nothing if no runs are seen.

    XBPM feedback isn't reliable during collections due to:
     * Objects (e.g. attenuator) crossing the beam can cause large (incorrect) feedback movements
     * Lower transmissions/higher energies are less reliable for the xbpm

    So we need to keep the transmission at 100% and the feedback on when not collecting
    and then turn it off and set the correct transmission for collection. The feedback
    mostly accounts for slow thermal drift so it is safe to assume that the beam is
    stable during a collection.

    Args:
        plan: The plan performing the data collection.
        devices (XBPMPauseDevices): Composite device including The XBPM device that is responsible for keeping
                                                        the beam in position, and attenuator
        desired_transmission_fraction (float): The desired transmission for the collection
        run_key_to_wrap: (str | None): Pausing XBPM and setting transmission is inserted after the 'open_run' message is seen with
        the matching run key, and unpausing and resetting transmission is inserted after the corresponding 'close_run' message is
        seen. If not specified, instead wrap the first run encountered.
    """

    _wrapped_run_name: None | str = None

    def head(msg: Msg):
        yield from check_and_pause_feedback(
            devices.xbpm_feedback,
            devices.attenuator,
            desired_transmission_fraction,
        )

        # Allow 'open_run' message to pass through
        yield msg

    def tail():
        yield from unpause_xbpm_feedback_and_set_transmission_to_1(
            devices.xbpm_feedback, devices.attenuator
        )

    def insert_plans(msg: Msg):
        # Wrap the specified run, or, if none specified, wrap the first run encountered
        nonlocal _wrapped_run_name

        match msg.command:
            case "open_run":
                # If we specified a run key, did we encounter it
                # If we didn't specify, then insert the plans and track the name of the run
                if (
                    not (run_key_to_wrap or _wrapped_run_name)
                    or run_key_to_wrap is msg.run
                ):
                    _wrapped_run_name = msg.run if msg.run else "unnamed_run"
                    return head(msg), None
            case "close_run":
                # Check if the run tracked from above was closed
                # An exception is raised in the RunEngine if two unnamed runs are opened
                # at the same time, so we are safe from unpausing on the wrong run
                if (_wrapped_run_name == "unnamed_run" and not msg.run) or (
                    msg.run and _wrapped_run_name and _wrapped_run_name is msg.run
                ):
                    return None, tail()

        return None, None

    # Contingency wrapper can cause unpausing to occur on exception and again on close_run.
    # Not needed after https://github.com/bluesky/bluesky/issues/1891
    return (
        yield from bpp.contingency_wrapper(
            plan_mutator(plan, insert_plans),
            except_plan=lambda _: unpause_xbpm_feedback_and_set_transmission_to_1(
                devices.xbpm_feedback,
                devices.attenuator,
            ),
        )
    )


transmission_and_xbpm_feedback_for_collection_decorator = make_decorator(
    transmission_and_xbpm_feedback_for_collection_wrapper
)



from unittest.mock import MagicMock, patch

import pytest
from bluesky.run_engine import RunEngine

from mx_bluesky.common.external_interaction.ispyb.exp_eye_store import BLSampleStatus
from mx_bluesky.common.plans.write_sample_status import (
    SampleStatusExceptionType,
    deposit_loaded_sample,
    deposit_sample_error,
)
from mx_bluesky.common.utils.exceptions import SampleException

TEST_SAMPLE_ID = 123456


@pytest.mark.parametrize(
    "exception_type, expected_sample_status, expected_raised_exception",
    [
        [
            SampleStatusExceptionType.BEAMLINE,
            BLSampleStatus.ERROR_BEAMLINE,
            AssertionError,
        ],
        [
            SampleStatusExceptionType.SAMPLE,
            BLSampleStatus.ERROR_SAMPLE,
            SampleException,
        ],
    ],
)
def test_depositing_sample_error_with_sample_or_beamline_exception(
    RE: RunEngine,
    exception_type: SampleStatusExceptionType,
    expected_sample_status: BLSampleStatus,
    expected_raised_exception: type,
):
    mock_expeye = MagicMock()
    with (
        patch(
            "mx_bluesky.common.external_interaction.callbacks.sample_handling.sample_handling_callback"
            ".ExpeyeInteraction",
            return_value=mock_expeye,
        ),
        pytest.raises(expected_raised_exception),
    ):
        RE(deposit_sample_error(exception_type, TEST_SAMPLE_ID))
    mock_expeye.update_sample_status.assert_called_once_with(
        TEST_SAMPLE_ID, expected_sample_status
    )


def test_depositing_sample_loaded(
    RE: RunEngine,
):
    mock_expeye = MagicMock()
    with patch(
        "mx_bluesky.common.external_interaction.callbacks.sample_handling.sample_handling_callback"
        ".ExpeyeInteraction",
        return_value=mock_expeye,
    ):
        RE(deposit_loaded_sample(TEST_SAMPLE_ID))
        mock_expeye.update_sample_status.assert_called_once_with(
            TEST_SAMPLE_ID, BLSampleStatus.LOADED
        )



#!/usr/bin/env python3
import argparse
import locale
import os
import re
import subprocess
from functools import partial
from sys import stderr, stdout

SETUP_CFG_PATTERN = re.compile("(.*?\\S)\\s*(@(.*))?$")
SETUP_UNPINNED_PATTERN = re.compile("(.*?\\S)\\s*([<>=]+(.*))?$")
PIP = "pip"


def rename_original(suffix):
    os.rename("setup.cfg", "setup.cfg" + suffix)


def normalize(package_name: str):
    return re.sub(r"[-_.]+", "-", package_name).lower()


def fetch_pin_versions() -> dict[str, str]:
    process = run_pip_freeze()
    if process.returncode == 0:
        output = process.stdout
        lines = output.split("\n")
        pin_versions = {}
        for line in lines:
            kvpair = line.split("==")
            if len(kvpair) != 2:
                stderr.write(f"Unable to parse {line} - ignored\n")
            else:
                pin_versions[normalize(kvpair[0]).strip()] = kvpair[1].strip()
        return pin_versions
    else:
        stderr.write(f"pip freeze failed with error code {process.returncode}\n")
        stderr.write(process.stderr)
        exit(1)


def run_pip_freeze():
    process = subprocess.run(
        [PIP, "freeze"], capture_output=True, encoding=locale.getpreferredencoding()
    )
    return process


def process_setup_cfg(input_fname, output_fname, dependency_processor):
    with open(input_fname) as input_file:
        with open(output_fname, "w") as output_file:
            process_files(input_file, output_file, dependency_processor)


def process_files(input_file, output_file, dependency_processor):
    while line := input_file.readline():
        output_file.write(line)
        if line.startswith("install_requires"):
            break
    while (line := input_file.readline()) and not line.startswith("["):
        if line.isspace():
            output_file.write(line)
        else:
            dependency_processor(line, output_file)
    output_file.write(line)
    while line := input_file.readline():
        output_file.write(line)


def strip_comment(line: str):
    split = line.rstrip("\n").split("#", 1)
    return split[0], (split[1] if len(split) > 1 else None)


def write_with_comment(comment, text, output_file):
    output_file.write(text)
    if comment:
        output_file.write(" #" + comment)
    output_file.write("\n")


def update_setup_cfg_line(version_map: dict[str, str], line, output_file):
    stripped_line, comment = strip_comment(line)
    if match := SETUP_UNPINNED_PATTERN.match(stripped_line):
        normalized_name = normalize(match[1].strip())
        if normalized_name not in version_map:
            stderr.write(
                f"Unable to find {normalized_name} in installed python packages\n"
            )
            exit(1)

        write_with_comment(
            comment,
            f"    {normalized_name} == {version_map[normalized_name]}",
            output_file,
        )
    else:
        output_file.write(line)


def write_commit_message(pinned_versions: dict[str, str]):
    message = f"Pin dependencies prior to release. Dodal {pinned_versions['dls-dodal']}, nexgen {pinned_versions['nexgen']}"
    stdout.write(message)


def unpin_versions(line, output_file):
    stripped_line, comment = strip_comment(line)
    if match := SETUP_CFG_PATTERN.match(stripped_line):
        if match[3] and match[3].strip().startswith("git+"):
            write_with_comment(comment, match[1], output_file)
            return

    output_file.write(line)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Pin dependency versions in setup.cfg")
    parser.add_argument(
        "--unpin",
        help="remove pinned hashes from setup.cfg prior to pip installing latest",
        action="store_true",
    )
    args = parser.parse_args()

    if args.unpin:
        rename_original(".orig")
        process_setup_cfg("setup.cfg.orig", "setup.cfg", unpin_versions)
    else:
        rename_original(".unpinned")
        installed_versions = fetch_pin_versions()
        process_setup_cfg(
            "setup.cfg.unpinned",
            "setup.cfg",
            partial(update_setup_cfg_line, installed_versions),
        )
        write_commit_message(installed_versions)



"""
Cleaner abstractions of the PV table.

Takes the PV tables from I24's setup_beamline and wraps a slightly more
abstract wrapper around them.
"""

from mx_bluesky.beamlines.i24.serial.setup_beamline import pv


class Pilatus:
    id = 58
    name = "pilatus"

    # fast, slow / width, height
    image_size_pixels = (2463, 2527)
    pixel_size_mm = (0.172, 0.172)
    image_size_mm = tuple(
        round(a * b, 3) for a, b in zip(image_size_pixels, pixel_size_mm, strict=False)
    )

    det_y_threshold = 560.0
    det_y_target = 566.0

    class pv:
        detector_distance = pv.pilat_detdist
        wavelength = pv.pilat_wavelength
        transmission = pv.pilat_filtertrasm
        file_name = pv.pilat_filename
        file_path = pv.pilat_filepath
        file_template = pv.pilat_filetemplate
        file_number = pv.pilat_filenumber
        beamx = pv.pilat_beamx
        beamy = pv.pilat_beamy

    def __str__(self) -> str:
        return self.name


class Eiger:
    id = 94
    name = "eiger"

    pixel_size_mm = (0.075, 0.075)
    image_size_pixels = (3108, 3262)

    image_size_mm = tuple(
        round(a * b, 3) for a, b in zip(image_size_pixels, pixel_size_mm, strict=False)
    )

    det_y_threshold = -10.0
    det_y_target = -22.0

    class pv:
        detector_distance = pv.eiger_detdist
        wavelength = pv.eiger_wavelength
        transmission = "BL24I-EA-PILAT-01:cam1:FilterTransm"
        filenameRBV = pv.eiger_ODfilenameRBV
        file_name = pv.eiger_ODfilename
        file_path = pv.eiger_ODfilepath
        file_template = None
        sequence_id = pv.eiger_seqID
        beamx = pv.eiger_beamx
        beamy = pv.eiger_beamy
        bit_depth = pv.eiger_bitdepthrbv

    def __str__(self) -> str:
        return self.name


Detector = Pilatus | Eiger



"""This module contains the experimental plans which hyperion can run.

The __all__ list in here are the plans that are externally available from outside Hyperion.
"""

from mx_bluesky.hyperion.experiment_plans.flyscan_xray_centre_plan import (
    flyscan_xray_centre,
)
from mx_bluesky.hyperion.experiment_plans.grid_detect_then_xray_centre_plan import (
    grid_detect_then_xray_centre,
)
from mx_bluesky.hyperion.experiment_plans.load_centre_collect_full_plan import (
    load_centre_collect_full,
)
from mx_bluesky.hyperion.experiment_plans.pin_centre_then_xray_centre_plan import (
    pin_tip_centre_then_xray_centre,
)
from mx_bluesky.hyperion.experiment_plans.rotation_scan_plan import (
    multi_rotation_scan,
    rotation_scan,
)

__all__ = [
    "flyscan_xray_centre",
    "grid_detect_then_xray_centre",
    "rotation_scan",
    "pin_tip_centre_then_xray_centre",
    "multi_rotation_scan",
    "load_centre_collect_full",
]



import math
from math import asin

from scanspec.core import AxesPoints, Axis
from scipy.constants import physical_constants

hc_in_eV_and_Angstrom: float = (
    physical_constants["speed of light in vacuum"][0]
    * physical_constants["Planck constant in eV/Hz"][0]
    * 1e10  # Angstroms per metre
)


def interconvert_eV_Angstrom(wavelength_or_energy: float) -> float:
    return hc_in_eV_and_Angstrom / wavelength_or_energy


def convert_eV_to_angstrom(hv: float) -> float:
    return interconvert_eV_Angstrom(hv)


def convert_angstrom_to_eV(wavelength: float) -> float:
    return interconvert_eV_Angstrom(wavelength)


def number_of_frames_from_scan_spec(scan_points: AxesPoints[Axis]):
    ax = list(scan_points.keys())[0]
    return len(scan_points[ax])


def energy_to_bragg_angle(energy_kev: float, d_a: float) -> float:
    """Compute the bragg angle given the energy in kev.

    Args:
        energy_kev:  The energy in keV
        d_a:         The lattice spacing in Angstroms
    Returns:
        The bragg angle in degrees
    """
    wavelength_a = convert_eV_to_angstrom(energy_kev * 1000)
    d = d_a
    return asin(wavelength_a / (2 * d)) * 180 / math.pi



import numpy as np
import pytest
from numpy.typing import DTypeLike
from scanspec.core import AxesPoints

from mx_bluesky.common.external_interaction.nexus.nexus_utils import (
    AxisDirection,
    create_goniometer_axes,
    vds_type_based_on_bit_depth,
)


@pytest.mark.parametrize(
    "bit_depth,expected_type",
    [(8, np.uint8), (16, np.uint16), (32, np.uint32), (100, np.uint16)],
)
def test_vds_type_is_expected_based_on_bit_depth(
    bit_depth: int, expected_type: DTypeLike
):
    assert vds_type_based_on_bit_depth(bit_depth) == expected_type


@pytest.fixture
def scan_points(test_rotation_params) -> AxesPoints:
    return test_rotation_params.scan_points


@pytest.mark.parametrize(
    "omega_axis_direction, expected_axis_direction",
    [[AxisDirection.NEGATIVE, -1], [AxisDirection.POSITIVE, 1]],
)
def test_omega_axis_direction_determined_from_features(
    omega_axis_direction: AxisDirection,
    expected_axis_direction: float,
    scan_points: AxesPoints,
):
    omega_start = 0
    gonio = create_goniometer_axes(
        omega_start, scan_points, (0, 0, 0), 0, 0, omega_axis_direction
    )
    assert gonio.axes_list[0].name == "omega" and gonio.axes_list[0].vector == (
        expected_axis_direction,
        0,
        0,
    )



"""Provides external interaction functionality to Hyperion, including Nexus file
creation, ISPyB deposition, and Zocalo processing submissions.

Functionality from this module can/should be used through the callback functions in
external_interaction.callbacks which can subscribe to the Bluesky RunEngine and handle
these various interactions based on the documents emitted by the RunEngine during the
execution of the experimental plan. It's not recommended to use the interaction classes
here directly in plans except through the use of such callbacks.
"""



from __future__ import annotations

import json
import shutil
from collections.abc import Callable, Sequence
from itertools import dropwhile, takewhile
from math import ceil
from typing import Any
from unittest.mock import MagicMock, call, patch

import h5py
import numpy as np
import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from dodal.devices.oav.oav_parameters import OAVParameters
from dodal.devices.synchrotron import SynchrotronMode
from dodal.devices.xbpm_feedback import Pause
from ophyd.status import Status
from ophyd_async.testing import set_mock_value

from mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback import (
    ZocaloCallback,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import (
    IspybIds,
    StoreInIspyb,
)
from mx_bluesky.common.external_interaction.nexus.nexus_utils import AxisDirection
from mx_bluesky.common.utils.exceptions import ISPyBDepositionNotMade
from mx_bluesky.hyperion.experiment_plans.rotation_scan_plan import (
    RotationScanComposite,
    calculate_motion_profile,
    multi_rotation_scan,
)
from mx_bluesky.hyperion.external_interaction.callbacks.__main__ import (
    create_rotation_callbacks,
)
from mx_bluesky.hyperion.external_interaction.callbacks.rotation.ispyb_callback import (
    RotationISPyBCallback,
)
from mx_bluesky.hyperion.external_interaction.callbacks.rotation.nexus_callback import (
    RotationNexusFileCallback,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.rotation import MultiRotationScan, RotationScan

from ....conftest import (
    DocumentCapturer,
    extract_metafile,
    fake_read,
    mx_acquisition_from_conn,
    raw_params_from_file,
    remap_upsert_columns,
)
from ..external_interaction.conftest import *  # noqa # for fixtures

TEST_OFFSET = 1
TEST_SHUTTER_OPENING_DEGREES = 2.5


def test_multi_rotation_scan_params():
    raw_params = raw_params_from_file(
        "tests/test_data/parameter_json_files/good_test_multi_rotation_scan_parameters.json"
    )
    params = MultiRotationScan(**raw_params)
    omega_starts = [s["omega_start_deg"] for s in raw_params["rotation_scans"]]
    for i, scan in enumerate(params.single_rotation_scans):
        assert scan.omega_start_deg == omega_starts[i]
        assert scan.nexus_vds_start_img == params.scan_indices[i]
        assert params.scan_indices

    detector_params = params.detector_params
    # MX-bluesky 563 assumptions are made about DetectorParams which aren't true for this test file
    assert detector_params.num_images_per_trigger == 1800
    assert detector_params.num_triggers == 3
    assert detector_params.omega_start == 0


async def test_multi_rotation_plan_runs_multiple_plans_in_one_arm(
    fake_create_rotation_devices: RotationScanComposite,
    test_multi_rotation_params: MultiRotationScan,
    sim_run_engine_for_rotation: RunEngineSimulator,
    oav_parameters_for_rotation: OAVParameters,
):
    smargon = fake_create_rotation_devices.smargon
    omega = smargon.omega
    set_mock_value(
        fake_create_rotation_devices.synchrotron.synchrotron_mode, SynchrotronMode.USER
    )
    msgs = sim_run_engine_for_rotation.simulate_plan(
        multi_rotation_scan(
            fake_create_rotation_devices,
            test_multi_rotation_params,
            oav_parameters_for_rotation,
        )
    )

    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "set" and msg.obj.name == "eiger_do_arm"
    )[1:]

    msgs_within_arming = list(
        takewhile(
            lambda msg: msg.command != "unstage"
            and (not msg.obj or msg.obj.name != "eiger"),
            msgs,
        )
    )

    def _assert_set_seq_and_return_remaining(remaining, name_value_pairs):
        for name, value in name_value_pairs:
            try:
                remaining = assert_message_and_return_remaining(
                    remaining,
                    lambda msg: msg.command == "set"
                    and msg.obj.name == name
                    and msg.args == (value,),
                )
            except Exception as e:
                raise Exception(f"Failed to find {name} being set to {value}") from e
        return remaining

    for scan in test_multi_rotation_params.single_rotation_scans:
        motion_values = calculate_motion_profile(
            scan,
            (await omega.acceleration_time.get_value()),
            (await omega.max_velocity.get_value()),
        )
        # moving to the start position
        msgs_within_arming = _assert_set_seq_and_return_remaining(
            msgs_within_arming,
            [
                ("smargon-x", scan.x_start_um / 1000),  # type: ignore
                ("smargon-y", scan.y_start_um / 1000),  # type: ignore
                ("smargon-z", scan.z_start_um / 1000),  # type: ignore
                ("smargon-phi", scan.phi_start_deg),
                ("smargon-chi", scan.chi_start_deg),
            ],
        )
        # arming the zebra
        msgs_within_arming = assert_message_and_return_remaining(
            msgs_within_arming,
            lambda msg: msg.command == "set" and msg.obj.name == "zebra-pc-arm",
        )
        # the final rel_set of omega to trigger the scan
        assert_message_and_return_remaining(
            msgs_within_arming,
            lambda msg: msg.command == "set"
            and msg.obj.name == "smargon-omega"
            and msg.args
            == (
                (scan.scan_width_deg + motion_values.shutter_opening_deg)
                * motion_values.direction.multiplier,
            ),
        )


def _run_multi_rotation_plan(
    RE: RunEngine,
    params: MultiRotationScan,
    devices: RotationScanComposite,
    callbacks: Sequence[Callable[[str, dict[str, Any]], Any]],
    oav_params: OAVParameters,
):
    for cb in callbacks:
        RE.subscribe(cb)
    with patch("bluesky.preprocessors.__read_and_stash_a_motor", fake_read):
        RE(multi_rotation_scan(devices, params, oav_params))


@patch(
    "mx_bluesky.hyperion.experiment_plans.rotation_scan_plan.check_topup_and_wait_if_necessary",
    autospec=True,
)
def test_full_multi_rotation_plan_docs_emitted(
    _,
    RE: RunEngine,
    test_multi_rotation_params: MultiRotationScan,
    fake_create_rotation_devices: RotationScanComposite,
    oav_parameters_for_rotation: OAVParameters,
):
    callback_sim = DocumentCapturer()
    _run_multi_rotation_plan(
        RE,
        test_multi_rotation_params,
        fake_create_rotation_devices,
        [callback_sim],
        oav_parameters_for_rotation,
    )
    docs = callback_sim.docs_received

    assert (
        outer_plan_start_doc := DocumentCapturer.assert_doc(
            docs, "start", matches_fields=({"plan_name": "multi_rotation_scan"})
        )
    )
    outer_uid = outer_plan_start_doc[1]["uid"]
    inner_run_docs = DocumentCapturer.get_docs_until(
        docs,
        "stop",
        matches_fields=({"run_start": outer_uid, "exit_status": "success"}),
    )[1:-1]

    for scan in test_multi_rotation_params.single_rotation_scans:
        inner_run_docs = DocumentCapturer.get_docs_from(
            inner_run_docs,
            "start",
            matches_fields={"subplan_name": "rotation_scan_with_cleanup"},
        )
        scan_docs = DocumentCapturer.get_docs_until(
            inner_run_docs,
            "stop",
            matches_fields={"run_start": inner_run_docs[0][1]["uid"]},
        )
        params = RotationScan(**json.loads(scan_docs[0][1]["mx_bluesky_parameters"]))
        assert params == scan
        assert len(events := DocumentCapturer.get_matches(scan_docs, "event")) == 3
        DocumentCapturer.assert_events_and_data_in_order(
            events,
            [
                ["eiger_odin_file_writer_id"],
                ["undulator-current_gap", "synchrotron-synchrotron_mode", "smargon-x"],
                [
                    "attenuator-actual_transmission",
                    "flux-flux_reading",
                    "dcm-energy_in_kev",
                    "eiger_bit_depth",
                ],
            ],
        )
        inner_run_docs = DocumentCapturer.get_docs_from(
            inner_run_docs,
            "stop",
            matches_fields={"run_start": inner_run_docs[0][1]["uid"]},
        )


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.rotation.nexus_callback.NexusWriter"
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.rotation_scan_plan.check_topup_and_wait_if_necessary",
    autospec=True,
)
def test_full_multi_rotation_plan_nexus_writer_called_correctly(
    _,
    mock_nexus_writer: MagicMock,
    RE: RunEngine,
    test_multi_rotation_params: MultiRotationScan,
    fake_create_rotation_devices: RotationScanComposite,
    oav_parameters_for_rotation: OAVParameters,
):
    callback = RotationNexusFileCallback()
    _run_multi_rotation_plan(
        RE,
        test_multi_rotation_params,
        fake_create_rotation_devices,
        [callback],
        oav_parameters_for_rotation,
    )
    nexus_writer_calls = mock_nexus_writer.call_args_list
    first_run_number = test_multi_rotation_params.detector_params.run_number
    for writer_call, rotation_params in zip(
        nexus_writer_calls,
        test_multi_rotation_params.single_rotation_scans,
        strict=False,
    ):
        callback_params = writer_call.args[0]
        assert callback_params == rotation_params
        assert writer_call.kwargs == {
            "omega_start_deg": rotation_params.omega_start_deg,
            "chi_start_deg": rotation_params.chi_start_deg,
            "phi_start_deg": rotation_params.phi_start_deg,
            "vds_start_index": rotation_params.nexus_vds_start_img,
            "full_num_of_images": test_multi_rotation_params.num_images,
            "meta_data_run_number": first_run_number,
            "axis_direction": AxisDirection.NEGATIVE
            if rotation_params.features.omega_flip
            else AxisDirection.POSITIVE,
        }


@patch(
    "mx_bluesky.hyperion.experiment_plans.rotation_scan_plan.check_topup_and_wait_if_necessary",
    autospec=True,
)
def test_full_multi_rotation_plan_nexus_files_written_correctly(
    _,
    RE: RunEngine,
    feature_flags_update_with_omega_flip: MagicMock,
    test_multi_rotation_params: MultiRotationScan,
    fake_create_rotation_devices: RotationScanComposite,
    oav_parameters_for_rotation: OAVParameters,
    tmpdir,
):
    multi_params = test_multi_rotation_params
    prefix = "multi_rotation_test"
    test_data_dir = "tests/test_data/nexus_files/"
    meta_file = f"{test_data_dir}rotation/ins_8_5_meta.h5.gz"
    fake_datafile = f"{test_data_dir}fake_data.h5"
    multi_params.file_name = prefix
    multi_params.storage_directory = f"{tmpdir}"
    meta_data_run_number = multi_params.detector_params.run_number

    data_filename_prefix = f"{prefix}_{meta_data_run_number}_"
    meta_filename = f"{prefix}_{meta_data_run_number}_meta.h5"

    callback = RotationNexusFileCallback()
    _run_multi_rotation_plan(
        RE,
        multi_params,
        fake_create_rotation_devices,
        [callback],
        oav_parameters_for_rotation,
    )

    def _expected_dset_number(image_number: int):
        # image numbers 0-999 are in dset 1, etc.
        return int(ceil((image_number + 1) / 1000))

    num_datasets = range(
        1, _expected_dset_number(multi_params.num_images - 1)
    )  # the index of the last image is num_images - 1

    for i in num_datasets:
        shutil.copy(
            fake_datafile,
            f"{tmpdir}/{data_filename_prefix}{i:06d}.h5",
        )
    extract_metafile(
        meta_file,
        f"{tmpdir}/{meta_filename}",
    )
    for i, scan in enumerate(multi_params.single_rotation_scans):
        with h5py.File(f"{tmpdir}/{prefix}_{i + 1}.nxs", "r") as written_nexus_file:
            # check links go to the right file:
            detector_specific = written_nexus_file[
                "entry/instrument/detector/detectorSpecific"
            ]
            for field in ["software_version"]:
                link = detector_specific.get(field, getlink=True)  # type: ignore
                assert link.filename == meta_filename  # type: ignore
            data_group = written_nexus_file["entry/data"]
            for field in [f"data_{n:06d}" for n in num_datasets]:
                link = data_group.get(field, getlink=True)  # type: ignore
                assert link.filename.startswith(data_filename_prefix)  # type: ignore

            # check dataset starts and stops are correct:
            assert isinstance(dataset := data_group["data"], h5py.Dataset)  # type: ignore
            assert dataset.is_virtual
            assert dataset[scan.num_images - 1, 0, 0] == 0
            with pytest.raises(IndexError):
                assert dataset[scan.num_images, 0, 0] == 0
            dataset_sources = dataset.virtual_sources()
            expected_dset_start = _expected_dset_number(multi_params.scan_indices[i])
            expected_dset_end = _expected_dset_number(multi_params.scan_indices[i + 1])
            dset_start_name = dataset_sources[0].dset_name
            dset_end_name = dataset_sources[-1].dset_name
            assert dset_start_name.endswith(f"data_{expected_dset_start:06d}")
            assert dset_end_name.endswith(f"data_{expected_dset_end:06d}")

            # check scan values are correct for each file:
            assert isinstance(
                chi := written_nexus_file["/entry/sample/sample_chi/chi"], h5py.Dataset
            )
            assert chi[:] == scan.chi_start_deg
            assert isinstance(
                phi := written_nexus_file["/entry/sample/sample_phi/phi"], h5py.Dataset
            )
            assert phi[:] == scan.phi_start_deg
            assert isinstance(
                omega := written_nexus_file["/entry/sample/sample_omega/omega"],
                h5py.Dataset,
            )
            omega = omega[:]
            assert isinstance(
                omega_end := written_nexus_file["/entry/sample/sample_omega/omega_end"],
                h5py.Dataset,
            )
            omega_end = omega_end[:]
            assert len(omega) == scan.num_images
            expected_omega_starts = np.linspace(
                scan.omega_start_deg,
                scan.omega_start_deg
                + ((scan.num_images - 1) * multi_params.rotation_increment_deg),
                scan.num_images,
            )
            assert np.allclose(omega, expected_omega_starts)
            expected_omega_ends = (
                expected_omega_starts + multi_params.rotation_increment_deg
            )
            assert np.allclose(omega_end, expected_omega_ends)
            assert isinstance(
                omega_transform := written_nexus_file[
                    "/entry/sample/transformations/omega"
                ],
                h5py.Dataset,
            )
            assert isinstance(omega_vec := omega_transform.attrs["vector"], np.ndarray)
            omega_flip = (
                feature_flags_update_with_omega_flip.mock_calls[0].args[0].omega_flip
            )
            assert tuple(omega_vec) == (-1.0 if omega_flip else 1.0, 0, 0)


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.rotation.ispyb_callback.StoreInIspyb"
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.rotation_scan_plan.check_topup_and_wait_if_necessary",
    autospec=True,
)
def test_full_multi_rotation_plan_ispyb_called_correctly(
    _,
    mock_ispyb_store: MagicMock,
    RE: RunEngine,
    test_multi_rotation_params: MultiRotationScan,
    fake_create_rotation_devices: RotationScanComposite,
    oav_parameters_for_rotation: OAVParameters,
):
    callback = RotationISPyBCallback()
    mock_ispyb_store.return_value = MagicMock(spec=StoreInIspyb)
    _run_multi_rotation_plan(
        RE,
        test_multi_rotation_params,
        fake_create_rotation_devices,
        [callback],
        oav_parameters_for_rotation,
    )
    ispyb_calls = mock_ispyb_store.call_args_list
    for instantiation_call, ispyb_store_calls, _ in zip(
        ispyb_calls,
        [  # there should be 4 calls to the IspybStore per run
            mock_ispyb_store.return_value.method_calls[i * 4 : (i + 1) * 4]
            for i in range(len(test_multi_rotation_params.rotation_scans))
        ],
        test_multi_rotation_params.single_rotation_scans,
        strict=False,
    ):
        assert instantiation_call.args[0] == CONST.SIM.ISPYB_CONFIG
        assert ispyb_store_calls[0][0] == "begin_deposition"
        assert ispyb_store_calls[1][0] == "update_deposition"
        assert ispyb_store_calls[2][0] == "update_deposition"
        assert ispyb_store_calls[3][0] == "end_deposition"


@patch(
    "mx_bluesky.hyperion.experiment_plans.rotation_scan_plan.check_topup_and_wait_if_necessary",
    autospec=True,
)
def test_full_multi_rotation_plan_ispyb_interaction_end_to_end(
    _,
    mock_ispyb_conn_multiscan,
    RE: RunEngine,
    test_multi_rotation_params: MultiRotationScan,
    fake_create_rotation_devices: RotationScanComposite,
    oav_parameters_for_rotation: OAVParameters,
):
    number_of_scans = len(test_multi_rotation_params.rotation_scans)
    callback = RotationISPyBCallback()
    _run_multi_rotation_plan(
        RE,
        test_multi_rotation_params,
        fake_create_rotation_devices,
        [callback],
        oav_parameters_for_rotation,
    )
    mx = mx_acquisition_from_conn(mock_ispyb_conn_multiscan)
    assert mx.get_data_collection_group_params.call_count == number_of_scans
    assert mx.get_data_collection_params.call_count == number_of_scans * 4
    upsert_keys = mx.get_data_collection_params()
    for upsert_calls, rotation_params in zip(
        [  # there should be 4 datacollection upserts per scan
            mx.upsert_data_collection.call_args_list[i * 4 : (i + 1) * 4]
            for i in range(len(test_multi_rotation_params.rotation_scans))
        ],
        test_multi_rotation_params.single_rotation_scans,
        strict=False,
    ):
        first_upsert_data = remap_upsert_columns(upsert_keys, upsert_calls[0].args[0])
        assert (
            first_upsert_data["axisend"] - first_upsert_data["axisstart"]
            == rotation_params.scan_width_deg
        )
        assert first_upsert_data["nimages"] == rotation_params.num_images
        second_upsert_data = remap_upsert_columns(upsert_keys, upsert_calls[1].args[0])
        dc_id = second_upsert_data["id"]
        append_comment_call = next(
            dropwhile(
                lambda c: c.args[0] != dc_id,
                mx.update_data_collection_append_comments.mock_calls,
            )
        )
        comment = append_comment_call.args[1]
        assert comment.startswith("Sample position")
        position_string = f"{rotation_params.x_start_um:.0f}, {rotation_params.y_start_um:.0f}, {rotation_params.z_start_um:.0f}"
        assert position_string in comment
        third_upsert_data = remap_upsert_columns(upsert_keys, upsert_calls[2].args[0])
        assert third_upsert_data["resolution"] > 0  # resolution
        assert third_upsert_data["focalspotsizeatsamplex"] > 0  # beam size
        fourth_upsert_data = remap_upsert_columns(upsert_keys, upsert_calls[3].args[0])
        assert fourth_upsert_data["endtime"]  # timestamp
        assert fourth_upsert_data["runstatus"] == "DataCollection Successful"


@patch(
    "mx_bluesky.hyperion.experiment_plans.rotation_scan_plan.check_topup_and_wait_if_necessary",
    autospec=True,
)
def test_full_multi_rotation_plan_arms_eiger_asynchronously_and_disarms(
    _,
    RE: RunEngine,
    test_multi_rotation_params: MultiRotationScan,
    fake_create_rotation_devices: RotationScanComposite,
    oav_parameters_for_rotation: OAVParameters,
):
    eiger = fake_create_rotation_devices.eiger
    eiger.stage = MagicMock(return_value=Status(done=True, success=True))
    eiger.unstage = MagicMock(return_value=Status(done=True, success=True))
    eiger.do_arm.set = MagicMock(return_value=Status(done=True, success=True))

    _run_multi_rotation_plan(
        RE,
        test_multi_rotation_params,
        fake_create_rotation_devices,
        [],
        oav_parameters_for_rotation,
    )
    # Stage will arm the eiger synchonously
    eiger.stage.assert_not_called()

    eiger.do_arm.set.assert_called_once()
    eiger.unstage.assert_called_once()


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.rotation.ispyb_callback.StoreInIspyb"
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.rotation_scan_plan.check_topup_and_wait_if_necessary",
    autospec=True,
)
def test_zocalo_callback_end_only_gets_called_at_the_end_of_all_collections(
    _,
    mock_ispyb_store: MagicMock,
    RE: RunEngine,
    test_multi_rotation_params: MultiRotationScan,
    fake_create_rotation_devices: RotationScanComposite,
    oav_parameters_for_rotation: OAVParameters,
):
    """We must unstage the detector before we trigger zocalo so that we're sure we've
    finished writing data."""
    mock_ispyb_store.return_value = MagicMock(spec=StoreInIspyb)
    mock_ispyb_store.return_value.begin_deposition.return_value = IspybIds(
        data_collection_ids=(123,)
    )
    eiger = fake_create_rotation_devices.eiger
    parent_mock = MagicMock()
    parent_mock.eiger = MagicMock(return_value=Status(done=True, success=True))
    eiger.unstage = parent_mock.eiger_unstage
    _, ispyb_callback = create_rotation_callbacks()
    zocalo_callback = ispyb_callback.emit_cb
    assert isinstance(zocalo_callback, ZocaloCallback)
    zocalo_callback.zocalo_interactor = MagicMock()
    zocalo_callback.zocalo_interactor.run_end = parent_mock.run_end

    _run_multi_rotation_plan(
        RE,
        test_multi_rotation_params,
        fake_create_rotation_devices,
        [ispyb_callback],
        oav_parameters_for_rotation,
    )

    assert parent_mock.method_calls.count(call.run_end(123)) == len(
        test_multi_rotation_params.rotation_scans
    )
    assert parent_mock.method_calls[0] == call.eiger_unstage


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.rotation.ispyb_callback.StoreInIspyb"
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.rotation_scan_plan.check_topup_and_wait_if_necessary",
    autospec=True,
)
def test_zocalo_start_and_end_not_triggered_if_ispyb_ids_not_present(
    _,
    mock_ispyb_store: MagicMock,
    RE: RunEngine,
    test_multi_rotation_params: MultiRotationScan,
    fake_create_rotation_devices: RotationScanComposite,
    oav_parameters_for_rotation: OAVParameters,
):
    _, ispyb_callback = create_rotation_callbacks()
    zocalo_callback = ispyb_callback.emit_cb
    assert isinstance(zocalo_callback, ZocaloCallback)
    zocalo_callback.zocalo_interactor = (zocalo_trigger := MagicMock())

    ispyb_callback.ispyb = MagicMock(spec=StoreInIspyb)
    with pytest.raises(ISPyBDepositionNotMade):
        _run_multi_rotation_plan(
            RE,
            test_multi_rotation_params,
            fake_create_rotation_devices,
            [ispyb_callback],
            oav_parameters_for_rotation,
        )

    zocalo_trigger.run_start.assert_not_called()  # type: ignore


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.rotation.ispyb_callback.StoreInIspyb"
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.rotation_scan_plan.check_topup_and_wait_if_necessary",
    autospec=True,
)
def test_ispyb_triggered_before_zocalo(
    _,
    mock_ispyb_store: MagicMock,
    RE: RunEngine,
    test_multi_rotation_params: MultiRotationScan,
    fake_create_rotation_devices: RotationScanComposite,
    oav_parameters_for_rotation: OAVParameters,
):
    _, ispyb_callback = create_rotation_callbacks()
    parent_mock = MagicMock()

    mock_ispyb_store.return_value = MagicMock(spec=StoreInIspyb)
    mock_ispyb_store.return_value.begin_deposition = parent_mock.ispyb_begin
    mock_ispyb_store.return_value.begin_deposition.return_value = IspybIds(
        data_collection_ids=(123,)
    )

    zocalo_callback = ispyb_callback.emit_cb
    assert isinstance(zocalo_callback, ZocaloCallback)
    zocalo_callback.zocalo_interactor = MagicMock()
    zocalo_callback.zocalo_interactor.run_start = parent_mock.zocalo_start

    _run_multi_rotation_plan(
        RE,
        test_multi_rotation_params,
        fake_create_rotation_devices,
        [ispyb_callback],
        oav_parameters_for_rotation,
    )

    call_names = [call[0] for call in parent_mock.method_calls]

    assert "ispyb_begin" in call_names
    assert "zocalo_start" in call_names

    assert call_names.index("ispyb_begin") < call_names.index("zocalo_start")


@patch(
    "mx_bluesky.hyperion.external_interaction.callbacks.rotation.ispyb_callback.StoreInIspyb"
)
@patch(
    "mx_bluesky.hyperion.experiment_plans.rotation_scan_plan.check_topup_and_wait_if_necessary",
    autospec=True,
)
def test_zocalo_start_and_end_called_once_for_each_collection(
    _,
    mock_ispyb_store: MagicMock,
    RE: RunEngine,
    test_multi_rotation_params: MultiRotationScan,
    fake_create_rotation_devices: RotationScanComposite,
    oav_parameters_for_rotation: OAVParameters,
):
    _, ispyb_callback = create_rotation_callbacks()

    mock_ispyb_store.return_value = MagicMock(spec=StoreInIspyb)
    mock_ispyb_store.return_value.begin_deposition.return_value = IspybIds(
        data_collection_ids=(123,)
    )

    zocalo_callback = ispyb_callback.emit_cb
    assert isinstance(zocalo_callback, ZocaloCallback)
    zocalo_callback.zocalo_interactor = MagicMock()

    _run_multi_rotation_plan(
        RE,
        test_multi_rotation_params,
        fake_create_rotation_devices,
        [ispyb_callback],
        oav_parameters_for_rotation,
    )

    assert zocalo_callback.zocalo_interactor.run_start.call_count == len(
        test_multi_rotation_params.rotation_scans
    )
    assert zocalo_callback.zocalo_interactor.run_end.call_count == len(
        test_multi_rotation_params.rotation_scans
    )


def test_multi_rotation_scan_does_not_change_transmission_back_until_after_data_collected(
    fake_create_rotation_devices: RotationScanComposite,
    test_multi_rotation_params: MultiRotationScan,
    sim_run_engine_for_rotation: RunEngineSimulator,
    oav_parameters_for_rotation: OAVParameters,
):
    msgs = sim_run_engine_for_rotation.simulate_plan(
        multi_rotation_scan(
            fake_create_rotation_devices,
            test_multi_rotation_params,
            oav_parameters_for_rotation,
        )
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "unstage" and msg.obj.name == "eiger",
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "xbpm_feedback-pause_feedback"
        and msg.args[0] == Pause.RUN.value,
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "attenuator"
        and msg.args[0] == 1.0,
    )


def test_multi_rotation_scan_does_not_verify_undulator_gap_until_before_run(
    fake_create_rotation_devices: RotationScanComposite,
    test_multi_rotation_params: MultiRotationScan,
    sim_run_engine_for_rotation: RunEngineSimulator,
    oav_parameters_for_rotation: OAVParameters,
):
    msgs = sim_run_engine_for_rotation.simulate_plan(
        multi_rotation_scan(
            fake_create_rotation_devices,
            test_multi_rotation_params,
            oav_parameters_for_rotation,
        )
    )
    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "set" and msg.obj.name == "undulator"
    )
    msgs = assert_message_and_return_remaining(
        msgs, lambda msg: msg.command == "open_run"
    )



"""
Startup utilities for chip
"""

import string

from mx_bluesky.beamlines.i24.serial.fixed_target.ft_utils import ChipType
from mx_bluesky.beamlines.i24.serial.log import SSX_LOGGER, log_on_entry
from mx_bluesky.beamlines.i24.serial.parameters import get_chip_format


@log_on_entry
def fiducials(chip_type: int):
    fiducial_list: list | None = None
    if chip_type in [ChipType.Oxford, ChipType.OxfordInner, ChipType.Minichip]:
        fiducial_list = []
    elif chip_type == ChipType.Custom:
        # No fiducial for custom
        SSX_LOGGER.warning("No fiducials for custom chip")
    else:
        SSX_LOGGER.warning(f"Unknown chip_type, {chip_type}, in fiducials")
    return fiducial_list


def pathli(l_in=None, way="typewriter", reverse=False):
    if l_in is None:
        l_in = []
    if reverse is True:
        li = list(reversed(l_in))
    else:
        li = list(l_in)
    long_list = []
    if li:
        if way == "typewriter":
            for i in range(len(li) ** 2):
                long_list.append(li[i % len(li)])
        elif way == "snake":
            lr = list(reversed(li))
            for rep in range(len(li)):
                if rep % 2 == 0:
                    long_list += li
                else:
                    long_list += lr
        elif way == "snake53":
            lr = list(reversed(li))
            for rep in range(53):
                if rep % 2 == 0:
                    long_list += li
                else:
                    long_list += lr
        elif way == "expand":
            for entry in li:
                for _ in range(len(li)):
                    long_list.append(entry)
        elif way == "expand28":
            for entry in li:
                for _ in range(28):
                    long_list.append(entry)
        elif way == "expand25":
            for entry in li:
                for _ in range(25):
                    long_list.append(entry)
        else:
            SSX_LOGGER.warning(f"No known path, way =  {way}")
    else:
        SSX_LOGGER.warning("No list written")
    return long_list


def zippum(list_1_args, list_2_args):
    list_1, type_1, reverse_1 = list_1_args
    list_2, type_2, reverse_2 = list_2_args
    A_path = pathli(list_1, type_1, reverse_1)
    B_path = pathli(list_2, type_2, reverse_2)
    zipped_list = []
    for a, b in zip(A_path, B_path, strict=False):
        zipped_list.append(a + b)
    return zipped_list


def get_alphanumeric(chip_type: ChipType):
    cell_format = get_chip_format(chip_type)
    blk_num = cell_format.x_blocks
    wnd_num = cell_format.x_num_steps
    uppercase_list = list(string.ascii_uppercase)[:blk_num]
    lowercase_list = list(string.ascii_lowercase + string.ascii_uppercase + "0")[
        :wnd_num
    ]
    number_list = [str(x) for x in range(1, blk_num + 1)]

    block_list = zippum([uppercase_list, "expand", 0], [number_list, "typewriter", 0])
    window_list = zippum(
        [lowercase_list, "expand", 0], [lowercase_list, "typewriter", 0]
    )

    alphanumeric_list = []
    for block in block_list:
        for window in window_list:
            alphanumeric_list.append(block + "_" + window)
    SSX_LOGGER.info(f"Length of alphanumeric list = {len(alphanumeric_list)}")
    return alphanumeric_list



from collections.abc import Generator

from bluesky import plan_stubs as bps
from bluesky import preprocessors as bpp
from bluesky.utils import Msg
from dodal.devices.dcm import DCM
from dodal.devices.detector import (
    DetectorParams,
)
from dodal.devices.detector.detector_motion import DetectorMotion, ShutterState
from dodal.devices.eiger import EigerDetector
from dodal.devices.i03.beamstop import Beamstop

from mx_bluesky.hyperion.device_setup_plans.check_beamstop import check_beamstop
from mx_bluesky.hyperion.device_setup_plans.position_detector import (
    set_detector_z_position,
    set_shutter,
)


def fill_in_energy_if_not_supplied(dcm: DCM, detector_params: DetectorParams):
    if not detector_params.expected_energy_ev:
        actual_energy_ev = 1000 * (yield from bps.rd(dcm.energy_in_kev))
        detector_params.expected_energy_ev = actual_energy_ev
    return detector_params


def start_preparing_data_collection_then_do_plan(
    beamstop: Beamstop,
    eiger: EigerDetector,
    detector_motion: DetectorMotion,
    detector_distance_mm: float | None,
    plan_to_run: Generator[Msg, None, None],
    group="ready_for_data_collection",
) -> Generator[Msg, None, None]:
    """Starts preparing for the next data collection and then runs the
    given plan.

     Preparation consists of:
     * Arming the Eiger
     * Moving the detector to the specified position
     * Opening the detect shutter
     If the plan fails it will disarm the eiger.
    """

    def wrapped_plan():
        yield from bps.abs_set(eiger.do_arm, 1, group=group)  # type: ignore # Fix types in ophyd-async (https://github.com/DiamondLightSource/mx-bluesky/issues/855)
        if detector_distance_mm:
            yield from set_detector_z_position(
                detector_motion, detector_distance_mm, group
            )
        yield from set_shutter(detector_motion, ShutterState.OPEN, group)
        yield from plan_to_run

    yield from check_beamstop(beamstop)
    yield from bpp.contingency_wrapper(
        wrapped_plan(),
        except_plan=lambda e: (yield from bps.stop(eiger)),  # type: ignore # Fix types in ophyd-async (https://github.com/DiamondLightSource/mx-bluesky/issues/855)
    )



from unittest.mock import MagicMock, call

import pytest
from bluesky import plan_stubs as bps
from bluesky.protocols import Movable
from dodal.devices.zebra.zebra import (
    I03Axes,
    Zebra,
)
from dodal.devices.zebra.zebra_controlled_shutter import (
    ZebraShutter,
    ZebraShutterControl,
)

from mx_bluesky.hyperion.device_setup_plans.setup_zebra import (
    bluesky_retry,
    configure_zebra_and_shutter_for_auto_shutter,
    setup_zebra_for_gridscan,
    setup_zebra_for_panda_flyscan,
    setup_zebra_for_rotation,
    tidy_up_zebra_after_gridscan,
)


async def _get_shutter_input_2(zebra: Zebra):
    return (
        await zebra.logic_gates.and_gates[zebra.mapping.AND_GATE_FOR_AUTO_SHUTTER]
        .sources[2]
        .get_value()
    )


async def _get_shutter_input_1(zebra: Zebra):
    return (
        await zebra.logic_gates.and_gates[zebra.mapping.AND_GATE_FOR_AUTO_SHUTTER]
        .sources[1]
        .get_value()
    )


async def test_zebra_set_up_for_panda_gridscan(
    RE, zebra: Zebra, zebra_shutter: ZebraShutter
):
    RE(setup_zebra_for_panda_flyscan(zebra, zebra_shutter, wait=True))
    assert (
        await zebra.output.out_pvs[zebra.mapping.outputs.TTL_DETECTOR].get_value()
        == zebra.mapping.sources.IN1_TTL
    )
    assert (
        await zebra.output.out_pvs[zebra.mapping.outputs.TTL_PANDA].get_value()
        == zebra.mapping.sources.IN3_TTL
    )
    assert await zebra_shutter.control_mode.get_value() == ZebraShutterControl.AUTO
    assert await _get_shutter_input_2(zebra) == zebra.mapping.sources.IN4_TTL
    assert await _get_shutter_input_1(zebra) == zebra.mapping.sources.SOFT_IN1


async def test_zebra_set_up_for_gridscan(RE, zebra: Zebra, zebra_shutter: ZebraShutter):
    RE(setup_zebra_for_gridscan(zebra, zebra_shutter, wait=True))
    assert (
        await zebra.output.out_pvs[zebra.mapping.outputs.TTL_DETECTOR].get_value()
        == zebra.mapping.sources.IN3_TTL
    )
    assert await _get_shutter_input_2(zebra) == zebra.mapping.sources.IN4_TTL
    assert await zebra_shutter.control_mode.get_value() == ZebraShutterControl.AUTO
    assert await _get_shutter_input_1(zebra) == zebra.mapping.sources.SOFT_IN1


async def test_zebra_set_up_for_rotation(RE, zebra: Zebra, zebra_shutter: ZebraShutter):
    RE(setup_zebra_for_rotation(zebra, zebra_shutter, wait=True))
    assert await zebra.pc.gate_trigger.get_value() == I03Axes.OMEGA.value
    assert await zebra.pc.gate_width.get_value() == pytest.approx(360, 0.01)
    assert await zebra_shutter.control_mode.get_value() == ZebraShutterControl.AUTO
    assert await _get_shutter_input_1(zebra) == zebra.mapping.sources.SOFT_IN1


async def test_zebra_cleanup(RE, zebra: Zebra, zebra_shutter: ZebraShutter):
    RE(tidy_up_zebra_after_gridscan(zebra, zebra_shutter, wait=True))
    assert (
        await zebra.output.out_pvs[zebra.mapping.outputs.TTL_DETECTOR].get_value()
        == zebra.mapping.sources.PC_PULSE
    )
    assert await _get_shutter_input_2(zebra) == zebra.mapping.sources.PC_GATE


class MyException(Exception):
    pass


def test_when_first_try_fails_then_bluesky_retry_tries_again(RE, done_status):
    mock_device = MagicMock(spec=Movable)

    @bluesky_retry
    def my_plan(value):
        yield from bps.abs_set(mock_device, value)

    mock_device.set.side_effect = [MyException(), done_status]

    RE(my_plan(10))

    assert mock_device.set.mock_calls == [call(10), call(10)]


def test_when_all_tries_fail_then_bluesky_retry_throws_error(RE, done_status):
    mock_device = MagicMock(spec=Movable)

    @bluesky_retry
    def my_plan(value):
        yield from bps.abs_set(mock_device, value)

    exception_2 = MyException()
    mock_device.set.side_effect = [MyException(), exception_2]

    with pytest.raises(MyException) as e:
        RE(my_plan(10))

    assert e.value == exception_2


async def test_configure_zebra_and_shutter_for_auto(
    RE, zebra: Zebra, zebra_shutter: ZebraShutter
):
    RE(
        configure_zebra_and_shutter_for_auto_shutter(
            zebra, zebra_shutter, zebra.mapping.sources.IN4_TTL
        )
    )
    assert await zebra_shutter.control_mode.get_value() == ZebraShutterControl.AUTO
    assert await _get_shutter_input_1(zebra) == zebra.mapping.sources.SOFT_IN1
    assert await _get_shutter_input_2(zebra) == zebra.mapping.sources.IN4_TTL



import os
from time import sleep

import pytest
from requests import get

from mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping import (
    get_proposal_and_session_from_visit_string,
)
from mx_bluesky.common.external_interaction.ispyb.exp_eye_store import (
    BLSampleStatus,
    ExpeyeInteraction,
)

CONTAINER_ID = int(os.environ.get("ST_CONTAINER_ID", 288588))

SAMPLE_ID = int(os.environ.get("ST_SAMPLE_ID", 5289780))


@pytest.mark.system_test
@pytest.mark.parametrize(
    "message, expected_message",
    [
        ("Oh no!", "Oh no!"),
        (
            "Long message that will be truncated " + ("*" * 255),
            "Long message that will be truncated " + ("*" * 219),
        ),
    ],
)
def test_start_and_end_robot_load(message: str, expected_message: str):
    """To confirm this test is successful go to
    https://ispyb-test.diamond.ac.uk/dc/visit/cm37235-2 and see that data is added
    when it's run.
    """
    proposal, session = get_proposal_and_session_from_visit_string(
        os.environ.get("ST_VISIT", "cm37235-2")
    )
    BARCODE = "test_barcode"

    expeye = ExpeyeInteraction()

    robot_action_id = expeye.start_load(proposal, session, SAMPLE_ID, 40, 3)

    sleep(0.5)

    print(f"Created {robot_action_id}")

    test_folder = "/dls/i03/data/2024/cm37235-2/xtal_snapshots"
    oav_snapshot = test_folder + "/235855_load_after_0.0.png"
    webcam_snapshot = test_folder + "/235855_webcam.jpg"
    expeye.update_barcode_and_snapshots(
        robot_action_id, BARCODE, oav_snapshot, webcam_snapshot
    )

    sleep(0.5)

    expeye.end_load(robot_action_id, "fail", message)

    get_robot_data_url = f"{expeye._base_url}/robot-actions/{robot_action_id}"
    response = get(get_robot_data_url, auth=expeye._auth)

    assert response.ok
    response = response.json()
    assert response["robotActionId"] == robot_action_id
    assert response["status"] == "ERROR"
    assert response["sampleId"] == SAMPLE_ID
    assert response["sampleBarcode"] == BARCODE
    assert response["message"] == expected_message


@pytest.mark.system_test
def test_update_sample_updates_the_sample_status():
    sample_handling = ExpeyeInteraction()
    output_sample = sample_handling.update_sample_status(
        SAMPLE_ID, BLSampleStatus.ERROR_SAMPLE
    )
    expected_status = "ERROR - sample"
    assert output_sample.bl_sample_status == expected_status
    assert output_sample.container_id == CONTAINER_ID



from typing import Self, TypeVar

from pydantic import BaseModel, model_validator

from mx_bluesky.common.parameters.components import (
    MxBlueskyParameters,
    WithCentreSelection,
    WithSample,
    WithVisit,
)
from mx_bluesky.hyperion.parameters.components import WithHyperionUDCFeatures
from mx_bluesky.hyperion.parameters.robot_load import (
    RobotLoadThenCentre,
)
from mx_bluesky.hyperion.parameters.rotation import MultiRotationScan

T = TypeVar("T", bound=BaseModel)


def construct_from_values(parent_context: dict, child_dict: dict, t: type[T]) -> T:
    values = {k: v for k, v in parent_context.items() if not isinstance(v, dict)}
    values |= child_dict
    return t(**values)


class LoadCentreCollect(
    MxBlueskyParameters,
    WithVisit,
    WithSample,
    WithCentreSelection,
    WithHyperionUDCFeatures,
):
    """Experiment parameters to perform the combined robot load,
    pin-tip centre and rotation scan operations."""

    robot_load_then_centre: RobotLoadThenCentre
    multi_rotation_scan: MultiRotationScan

    @model_validator(mode="before")
    @classmethod
    def validate_model(cls, values):
        allowed_keys = (
            LoadCentreCollect.model_fields.keys()
            | RobotLoadThenCentre.model_fields.keys()
            | MultiRotationScan.model_fields.keys()
        )

        disallowed_keys = values.keys() - allowed_keys
        assert disallowed_keys == set(), (
            f"Unexpected fields found in LoadCentreCollect {disallowed_keys}"
        )

        keys_from_outer_load_centre_collect = (
            MxBlueskyParameters.model_fields.keys()
            | WithSample.model_fields.keys()
            | WithVisit.model_fields.keys()
        )
        duplicated_robot_load_then_centre_keys = (
            keys_from_outer_load_centre_collect
            & values["robot_load_then_centre"].keys()
        )
        assert not (duplicated_robot_load_then_centre_keys), (
            f"Unexpected keys in robot_load_then_centre: {', '.join(duplicated_robot_load_then_centre_keys)}"
        )

        duplicated_multi_rotation_scan_keys = (
            keys_from_outer_load_centre_collect & values["multi_rotation_scan"].keys()
        )
        assert not (duplicated_multi_rotation_scan_keys), (
            f"Unexpected keys in multi_rotation_scan: {', '.join(duplicated_multi_rotation_scan_keys)}"
        )

        new_robot_load_then_centre_params = construct_from_values(
            values, values["robot_load_then_centre"], RobotLoadThenCentre
        )
        new_multi_rotation_scan_params = construct_from_values(
            values, values["multi_rotation_scan"], MultiRotationScan
        )
        values["multi_rotation_scan"] = new_multi_rotation_scan_params
        values["robot_load_then_centre"] = new_robot_load_then_centre_params
        return values

    @model_validator(mode="after")
    def _check_rotation_start_xyz_is_not_specified(self) -> Self:
        for scan in self.multi_rotation_scan.single_rotation_scans:
            assert (
                not scan.x_start_um and not scan.y_start_um and not scan.z_start_um
            ), (
                "Specifying start xyz for sweeps is not supported in combination with centring."
            )
        return self

    @model_validator(mode="after")
    def _check_different_gridscan_and_rotation_energy_not_specified(self) -> Self:
        assert (
            self.multi_rotation_scan.demand_energy_ev is None
            or self.multi_rotation_scan.demand_energy_ev
            == self.robot_load_then_centre.demand_energy_ev
        ), "Setting a different energy for gridscan and rotation is not supported."
        return self



from bluesky import plan_stubs as bps
from bluesky import preprocessors as bpp
from dodal.devices.baton import Baton

from mx_bluesky.common.utils.exceptions import WarningException
from mx_bluesky.hyperion.experiment_plans.load_centre_collect_full_plan import (
    LoadCentreCollectComposite,
    load_centre_collect_full,
)
from mx_bluesky.hyperion.external_interaction.agamemnon import (
    create_parameters_from_agamemnon,
)
from mx_bluesky.hyperion.parameters.load_centre_collect import LoadCentreCollect

HYPERION_USER = "Hyperion"
NO_USER = "None"


def wait_for_hyperion_requested(baton: Baton):
    SLEEP_PER_CHECK = 0.1
    while True:
        requested_user = yield from bps.rd(baton.requested_user)
        if requested_user == HYPERION_USER:
            break
        yield from bps.sleep(SLEEP_PER_CHECK)


def ignore_sample_errors(exception: Exception):
    yield from bps.null()
    # For sample errors we want to continue the loop
    if not isinstance(exception, WarningException):
        raise exception


def main_hyperion_loop(baton: Baton, composite: LoadCentreCollectComposite):
    requested_user = yield from bps.rd(baton.requested_user)
    while requested_user == HYPERION_USER:

        def inner_loop():
            parameters: LoadCentreCollect | None = create_parameters_from_agamemnon()  # type: ignore # not complete until https://github.com/DiamondLightSource/mx-bluesky/issues/773
            if parameters:
                yield from load_centre_collect_full(composite, parameters)
            else:
                yield from bps.mv(baton.requested_user, NO_USER)

        yield from bpp.contingency_wrapper(
            inner_loop(), except_plan=ignore_sample_errors, auto_raise=False
        )
        requested_user = yield from bps.rd(baton.requested_user)


def move_to_default_state():
    # To be filled in in https://github.com/DiamondLightSource/mx-bluesky/issues/396
    yield from bps.null()


def run_udc_when_requested(baton: Baton, composite: LoadCentreCollectComposite):
    """This will wait for the baton to be handed to hyperion and then run through the
    UDC queue from agamemnon until:
      1. There are no more instructions from agamemnon
      2. There is an error on the beamline
      3. The baton is requested by another party

    In the case of 1. or 2. hyperion will immediately release the baton. In the case of
    3. the baton will be released after the next collection has finished."""

    yield from wait_for_hyperion_requested(baton)
    yield from bps.abs_set(baton.current_user, HYPERION_USER)

    def default_state_then_collect():
        yield from move_to_default_state()
        yield from main_hyperion_loop(baton, composite)

    def release_baton():
        # If hyperion has given up the baton itself we need to also release requested
        # user so that hyperion doesn't think we're requested again
        requested_user = yield from bps.rd(baton.requested_user)
        if requested_user == HYPERION_USER:
            yield from bps.abs_set(baton.requested_user, NO_USER)
        yield from bps.abs_set(baton.current_user, NO_USER)

    yield from bpp.contingency_wrapper(
        default_state_then_collect(), final_plan=release_baton
    )



import pytest

from mx_bluesky.hyperion.parameters.constants import CONST


@pytest.fixture
def test_rotation_start_outer_document(dummy_rotation_params):
    return {
        "uid": "d8bee3ee-f614-4e7a-a516-25d6b9e87ef3",
        "subplan_name": CONST.PLAN.ROTATION_OUTER,
        "mx_bluesky_parameters": dummy_rotation_params.model_dump_json(),
    }



import pytest

from mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_StartUp_py3v1 import (
    fiducials,
    pathli,
)


def test_fiducials():
    assert fiducials(0) == []
    assert fiducials(1) == []
    assert fiducials(2) is None


@pytest.mark.parametrize(
    "list_in, way, reverse, expected_res",
    [
        (
            [1, 2, 3],
            "typewriter",
            False,
            [1, 2, 3] * 3,
        ),  # Result should be list * len(list)
        ([1, 2, 3], "typewriter", True, [3, 2, 1] * 3),  # list[::-1] * len(list)
        ([4, 5], "snake", False, [4, 5, 5, 4]),  # Snakes the list
        ([4, 5], "expand", False, [4, 4, 5, 5]),  # Repeats each value
    ],
)
def test_pathli(list_in, way, reverse, expected_res):
    assert pathli(list_in, way, reverse) == expected_res



import pytest

from ....conftest import DocumentCapturer


@pytest.fixture
def test_docs():
    return [
        ("start", {"uid": 12345, "abc": 56789, "xyz": 99999}),
        ("stop", {"uid": 77777, "abc": 88888, "xyz": 99999}),
    ]


def test_callback_sim_doc_names(test_docs):
    DocumentCapturer.assert_doc(test_docs, "start")
    DocumentCapturer.assert_doc(test_docs, "stop")
    DocumentCapturer.assert_doc(test_docs, "restart", does_exist=False)


def test_callback_sim_has_fields(test_docs):
    DocumentCapturer.assert_doc(test_docs, "start", has_fields=["uid"])
    DocumentCapturer.assert_doc(test_docs, "stop", has_fields=["abc", "xyz"])
    DocumentCapturer.assert_doc(
        test_docs, "start", has_fields=["uid", "bbb"], does_exist=False
    )


def test_callback_sim_matches_fields(test_docs):
    DocumentCapturer.assert_doc(test_docs, "start", matches_fields={"uid": 12345})
    DocumentCapturer.assert_doc(
        test_docs, "stop", matches_fields={"abc": 88888, "xyz": 99999}
    )
    DocumentCapturer.assert_doc(
        test_docs,
        "start",
        matches_fields={"abc": 88888, "xyz": 99799},
        does_exist=False,
    )


def test_callback_sim_assert_switch(test_docs):
    with pytest.raises(AssertionError):
        DocumentCapturer.assert_doc(test_docs, "restart")

    with pytest.raises(AssertionError):
        DocumentCapturer.assert_doc(test_docs, "start", has_fields=["uid", "bbb"])

    with pytest.raises(AssertionError):
        DocumentCapturer.assert_doc(
            test_docs, "start", matches_fields={"abc": 88888, "xyz": 99799}
        )



from __future__ import annotations

import time
from datetime import UTC, datetime, timedelta
from enum import Enum

import numpy as np
from dodal.devices.detector import DetectorParams
from nexgen.nxs_utils import Attenuator, Axis, Beam, Detector, EigerDetector, Goniometer
from nexgen.nxs_utils.axes import TransformationType
from numpy.typing import DTypeLike

from mx_bluesky.common.utils.log import NEXUS_LOGGER
from mx_bluesky.common.utils.utils import convert_eV_to_angstrom


class AxisDirection(Enum):
    """
    Identifies whether the omega axis of rotation is on the positive x-axis or
    negative x-axis as per the Nexus standard:
    https://journals.iucr.org/m/issues/2020/05/00/ti5018/ti5018.pdf
    """

    POSITIVE = 1
    NEGATIVE = -1


def vds_type_based_on_bit_depth(detector_bit_depth: int) -> DTypeLike:
    """Works out the datatype for the VDS, based on the bit depth from the detector."""
    if detector_bit_depth == 8:
        return np.uint8
    elif detector_bit_depth == 16:
        return np.uint16
    elif detector_bit_depth == 32:
        return np.uint32
    else:
        NEXUS_LOGGER.error(
            f"Unknown detector bit depth {detector_bit_depth}, assuming 16-bit"
        )
        return np.uint16


def create_goniometer_axes(
    omega_start: float,
    scan_points: dict | None,
    x_y_z_increments: tuple[float, float, float] = (0.0, 0.0, 0.0),
    chi: float = 0.0,
    phi: float = 0.0,
    omega_axis_direction: AxisDirection = AxisDirection.NEGATIVE,
) -> Goniometer:
    """Returns a Nexgen 'Goniometer' object with the dependency chain of I03's Smargon
    goniometer. If scan points is provided these values will be used in preference to
    those from the params object.

    Args:
        omega_start (float): the starting position of omega, the only extra value that
                             needs to be specified except for the scan points.
        scan_points (dict):  a dictionary of points in the scan for each axis. Obtained
                             by calculating the scan path with scanspec and calling
                             consume() on it.
        x_y_z_increments:    optionally, specify the increments between each image for
                             the x, y, and z axes. Will be ignored if scan_points
                             is provided.
        omega_axis_direction:      The direction of the omega axis, this determines the
                             coordinate space parity
    Returns:
        The created Goniometer object
    """
    gonio_axes = [
        Axis(
            "omega",
            ".",
            TransformationType.ROTATION,
            (1.0 * omega_axis_direction.value, 0.0, 0.0),
            omega_start,
        ),
        Axis(
            name="sam_z",
            depends="omega",
            transformation_type=TransformationType.TRANSLATION,
            vector=(0.0, 0.0, 1.0),
            start_pos=0.0,
            increment=x_y_z_increments[2],
        ),
        Axis(
            name="sam_y",
            depends="sam_z",
            transformation_type=TransformationType.TRANSLATION,
            vector=(0.0, 1.0, 0.0),
            start_pos=0.0,
            increment=x_y_z_increments[1],
        ),
        Axis(
            name="sam_x",
            depends="sam_y",
            transformation_type=TransformationType.TRANSLATION,
            vector=(1.0, 0.0, 0.0),
            start_pos=0.0,
            increment=x_y_z_increments[0],
        ),
        Axis(
            "chi", "sam_x", TransformationType.ROTATION, (0.006, -0.0264, 0.9996), chi
        ),
        Axis("phi", "chi", TransformationType.ROTATION, (-1, -0.0025, -0.0056), phi),
    ]
    return Goniometer(gonio_axes, scan_points)


def get_start_and_predicted_end_time(time_expected: float) -> tuple[str, str]:
    time_format = r"%Y-%m-%dT%H:%M:%SZ"
    start = datetime.fromtimestamp(time.time(), tz=UTC)
    end_est = start + timedelta(seconds=time_expected)
    return start.strftime(time_format), end_est.strftime(time_format)


def create_detector_parameters(detector_params: DetectorParams) -> Detector:
    """Returns the detector information in a format that nexgen wants.

    Args:
        detector_params (DetectorParams): The detector params as Hyperion stores them.

    Returns:
        Detector: Detector description for nexgen.
    """
    detector_pixels = detector_params.get_detector_size_pizels()

    eiger_params = EigerDetector(
        "Eiger 16M", (detector_pixels.height, detector_pixels.width), "Si", 46051, 0
    )

    detector_axes = [
        Axis(
            "det_z",
            ".",
            TransformationType.TRANSLATION,
            (0.0, 0.0, 1.0),
            detector_params.detector_distance,
        )
    ]
    # Eiger parameters, axes, beam_center, exp_time, [fast, slow]
    return Detector(
        eiger_params,
        detector_axes,
        list(
            detector_params.get_beam_position_pixels(detector_params.detector_distance)
        ),
        detector_params.exposure_time_s,
        [(-1.0, 0.0, 0.0), (0.0, -1.0, 0.0)],
    )


def create_beam_and_attenuator_parameters(
    energy_kev: float, flux: float, transmission_fraction: float
) -> tuple[Beam, Attenuator]:
    """Create beam and attenuator objects that nexgen can understands

    Returns:
        tuple[Beam, Attenuator]: Descriptions of the beam and attenuator for nexgen.
    """
    return (
        Beam(convert_eV_to_angstrom(energy_kev * 1000), flux),  # pyright: ignore
        Attenuator(transmission_fraction),  # pyright: ignore
    )



import shutil
from pathlib import Path
from unittest.mock import Mock

import bluesky.plan_stubs as bps
import pytest
from bluesky.preprocessors import run_decorator, set_run_key_decorator
from bluesky.run_engine import RunEngine
from dodal.devices.oav.oav_detector import OAV
from PIL import Image

from mx_bluesky.common.parameters.constants import DocDescriptorNames
from mx_bluesky.hyperion.external_interaction.callbacks.snapshot_callback import (
    BeamDrawingCallback,
)
from mx_bluesky.hyperion.parameters.constants import CONST
from mx_bluesky.hyperion.parameters.rotation import RotationScan

from ......conftest import raw_params_from_file


@pytest.fixture
def params():
    return RotationScan(
        **raw_params_from_file(
            "tests/test_data/parameter_json_files/good_test_rotation_scan_parameters.json"
        )
    )


def simplified_experiment_plan(
    oav: OAV, snapshot_directory: Path, params: RotationScan
):
    @set_run_key_decorator(CONST.PLAN.LOAD_CENTRE_COLLECT)
    @run_decorator(
        md={
            "activate_callbacks": ["BeamDrawingCallback"],
            "with_snapshot": params.model_dump_json(),
        }
    )
    def inner():
        # pretend as if we triggered the oav snapshot
        yield from bps.abs_set(
            oav.snapshot.last_saved_path, str(snapshot_directory / "test_filename.png")
        )
        yield from bps.create(DocDescriptorNames.OAV_ROTATION_SNAPSHOT_TRIGGERED)
        yield from bps.read(oav)
        yield from bps.save()

    yield from inner()


def test_snapshot_callback_loads_and_saves_updated_snapshot_propagates_event(
    tmp_path: Path, RE: RunEngine, oav: OAV, params: RotationScan
):
    downstream_cb = Mock()
    callback = BeamDrawingCallback(emit=downstream_cb)
    base_image_path = tmp_path / "test_filename.png"
    shutil.copy(
        "tests/test_data/test_images/generate_snapshot_input.png", base_image_path
    )

    RE.subscribe(callback)
    RE(simplified_experiment_plan(oav, tmp_path, params))

    expected_image = Image.open(
        "tests/test_data/test_images/generate_snapshot_output.png"
    )
    expected_bytes = expected_image.tobytes()
    generated_image_path = str(tmp_path / "test_filename_with_beam_centre.png")
    generated_image = Image.open(generated_image_path)
    generated_bytes = generated_image.tobytes()
    assert generated_bytes == expected_bytes, "Actual and expected images differ"

    downstream_calls = downstream_cb.mock_calls
    assert downstream_calls[0].args[0] == "start"
    assert downstream_calls[1].args[0] == "descriptor"
    assert downstream_calls[2].args[0] == "event"
    assert (
        downstream_calls[2].args[1]["data"]["oav-snapshot-last_saved_path"]
        == generated_image_path
    )
    assert downstream_calls[3].args[0] == "stop"



import bluesky.plan_stubs as bps
from bluesky.utils import MsgGenerator
from dodal.devices.aithre_lasershaping.goniometer import Goniometer

DEFAULT_VELOCITIES = [5.0, 10.0, 20.0, 40.0, 80.0, 90.0]
DEFAULT_POSITIONS = [
    300.0,
    -300.0,
    600.0,
    -600.0,
    1200.0,
    -1200.0,
    2400.0,
    -2400.0,
    3600.0,
    -3600.0,
]


def check_omega_performance(
    goniometer: Goniometer,
    velocities: list[float] = DEFAULT_VELOCITIES,
    values: list[float] = DEFAULT_POSITIONS,
) -> MsgGenerator:
    """Move the goniometer from positive to negative to check omega performance"""
    for omega_velocity in velocities:
        yield from bps.abs_set(goniometer.omega.velocity, omega_velocity, wait=True)
        for omega_value in values:
            yield from bps.abs_set(goniometer.omega, omega_value, wait=True)



import os
from enum import Enum

from dodal.devices.aperturescatterguard import ApertureValue
from dodal.devices.detector import EIGER2_X_16M_SIZE
from dodal.devices.zocalo.zocalo_constants import ZOCALO_ENV as ZOCALO_ENV_FROM_DODAL
from dodal.utils import get_beamline_name
from pydantic.dataclasses import dataclass

from mx_bluesky.definitions import ROOT_DIR

BEAMLINE = get_beamline_name("test")
TEST_MODE = BEAMLINE == "test"


@dataclass(frozen=True)
class DocDescriptorNames:
    # Robot load event descriptor
    ROBOT_LOAD = "robot_load"
    # For callbacks to use
    OAV_ROTATION_SNAPSHOT_TRIGGERED = "rotation_snapshot_triggered"
    OAV_GRID_SNAPSHOT_TRIGGERED = "snapshot_to_ispyb"
    HARDWARE_READ_PRE = "read_hardware_for_callbacks_pre_collection"
    HARDWARE_READ_DURING = "read_hardware_for_callbacks_during_collection"
    SAMPLE_HANDLING_EXCEPTION = "sample_handling_exception"
    ZOCALO_HW_READ = "zocalo_read_hardware_plan"
    FLYSCAN_RESULTS = "flyscan_results_obtained"


@dataclass(frozen=True)
class OavConstants:
    OAV_CONFIG_JSON = (
        "tests/test_data/test_OAVCentring.json"
        if TEST_MODE
        else (
            f"/dls_sw/{BEAMLINE}/software/daq_configuration/json/OAVCentring_hyperion.json"
        )
    )


@dataclass(frozen=True)
class PlanNameConstants:
    LOAD_CENTRE_COLLECT = "load_centre_collect"
    # Robot load subplan
    ROBOT_LOAD = "robot_load"
    # Gridscan
    GRID_DETECT_AND_DO_GRIDSCAN = "grid_detect_and_do_gridscan"
    GRID_DETECT_INNER = "grid_detect"
    GRIDSCAN_OUTER = "run_gridscan_move_and_tidy"
    GRIDSCAN_AND_MOVE = "run_gridscan_and_move"
    GRIDSCAN_MAIN = "run_gridscan"
    DO_FGS = "do_fgs"
    # IspyB callback activation
    ISPYB_ACTIVATION = "ispyb_activation"
    ROBOT_LOAD_AND_SNAPSHOTS = "robot_load_and_snapshots"
    # Rotation scan
    ROTATION_MULTI = "multi_rotation_wrapper"
    ROTATION_OUTER = "rotation_scan_with_cleanup"
    ROTATION_MAIN = "rotation_scan_main"
    FLYSCAN_RESULTS = "xray_centre_results"
    SET_ENERGY = "set_energy"
    UNNAMED_RUN = "unnamed_run"


@dataclass(frozen=True)
class EnvironmentConstants:
    ZOCALO_ENV = ZOCALO_ENV_FROM_DODAL


@dataclass(frozen=True)
class TriggerConstants:
    ZOCALO = "trigger_zocalo_on"


@dataclass(frozen=True)
class HardwareConstants:
    OAV_REFRESH_DELAY = 0.3
    PANDA_FGS_RUN_UP_DEFAULT = 0.17
    CRYOJET_MARGIN_MM = 0.2
    THAWING_TIME = 20
    TIP_OFFSET_UM = 0

    # Value quoted in https://www.dectris.com/en/detectors/x-ray-detectors/eiger2/eiger2-for-synchrotrons/eiger2-x/,
    # causes dropped frames, so increase value for safety
    PANDA_FGS_EIGER_DEADTIME_S = 5e-5


@dataclass(frozen=True)
class GridscanParamConstants:
    WIDTH_UM = 600.0
    EXPOSURE_TIME_S = 0.004
    USE_ROI = True
    BOX_WIDTH_UM = 20.0
    OMEGA_1 = 0.0
    OMEGA_2 = 90.0
    PANDA_RUN_UP_DISTANCE_MM = 0.2


@dataclass(frozen=True)
class RotationParamConstants:
    DEFAULT_APERTURE_POSITION = ApertureValue.LARGE


@dataclass(frozen=True)
class DetectorParamConstants:
    BEAM_XY_LUT_PATH = (
        "tests/test_data/test_det_dist_converter.txt"
        if TEST_MODE
        else f"/dls_sw/{BEAMLINE}/software/daq_configuration/lookup/DetDistToBeamXYConverter.txt"
    )
    DETECTOR = EIGER2_X_16M_SIZE


@dataclass(frozen=True)
class ExperimentParamConstants:
    DETECTOR = DetectorParamConstants()
    GRIDSCAN = GridscanParamConstants()
    ROTATION = RotationParamConstants()


@dataclass(frozen=True)
class PlanGroupCheckpointConstants:
    # For places to synchronise / stop and wait in plans, use as bluesky group names
    GRID_READY_FOR_DC = "grid_ready_for_data_collection"
    ROTATION_READY_FOR_DC = "rotation_ready_for_data_collection"
    MOVE_GONIO_TO_START = "move_gonio_to_start"
    READY_FOR_OAV = "ready_for_oav"


# Eventually replace below with https://github.com/DiamondLightSource/mx-bluesky/issues/798
@dataclass(frozen=True)
class DeviceSettingsConstants:
    PANDA_FLYSCAN_SETTINGS_FILENAME = "panda-gridscan"
    PANDA_FLYSCAN_SETTINGS_DIR = os.path.abspath(
        f"{ROOT_DIR}/hyperion/resources/panda/"
    )


@dataclass(frozen=True)
class SimConstants:
    BEAMLINE = "BL03S"
    INSERTION_PREFIX = "SR03S"
    # this one is for unit tests
    ISPYB_CONFIG = "tests/test_data/test_config.cfg"
    # this one is for system tests
    DEV_ISPYB_DATABASE_CFG = "/dls_sw/dasc/mariadb/credentials/ispyb-hyperion-dev.cfg"


class Actions(Enum):
    START = "start"
    STOP = "stop"
    SHUTDOWN = "shutdown"
    STATUS = "status"


class Status(Enum):
    WARN = "Warn"
    FAILED = "Failed"
    SUCCESS = "Success"
    BUSY = "Busy"
    ABORTING = "Aborting"
    IDLE = "Idle"



from enum import StrEnum
from os import environ
from pathlib import Path

from mx_bluesky.beamlines.i24.serial.log import _read_visit_directory_from_file


class SSXType(StrEnum):
    FIXED = "Serial Fixed"
    EXTRUDER = "Serial Jet"


class DetectorName(StrEnum):
    EIGER = "eiger"
    PILATUS = "pilatus"


# TODO figue sth out for tests
LUT_FILES_PATH = Path("/dls_sw/i24/software/daq_configuration/lookup")

BEAM_CENTER_LUT_FILES = {
    DetectorName.EIGER: LUT_FILES_PATH / "DetDistToBeamXYConverterE9M.txt",
    DetectorName.PILATUS: LUT_FILES_PATH / "DetDistToBeamXYConverterP6M.txt",
}


OAV_CONFIG_FILES = {
    "zoom_params_file": "/dls_sw/i24/software/gda_versions/gda_9_36/config/xml/jCameraManZoomLevels.xml",
    "oav_config_json": "/dls_sw/i24/software/daq_configuration/json/OAVCentring.json",
    "display_config": "/dls_sw/i24/software/gda_versions/var/display.configuration",
}
OAV1_CAM = "http://bl24i-di-serv-01.diamond.ac.uk:8080/OAV1.mjpg.mjpg"

HEADER_FILES_PATH = Path("/dls_sw/i24/scripts/fastchips/").expanduser().resolve()

INTERNAL_FILES_PATH = Path(__file__).absolute().parent


def _params_file_location() -> Path:
    beamline: str | None = environ.get("BEAMLINE")
    filepath: Path

    if beamline:
        filepath = _read_visit_directory_from_file() / "tmp/serial/parameters"
    else:
        filepath = INTERNAL_FILES_PATH

    filepath.mkdir(parents=True, exist_ok=True)

    return filepath


PARAM_FILE_NAME = "parameters.json"
# Paths for rw - these should have been created on startup
PARAM_FILE_PATH = _params_file_location()
PARAM_FILE_PATH_FT = PARAM_FILE_PATH / "fixed_target"
LITEMAP_PATH = PARAM_FILE_PATH_FT / "litemaps"
# Paths for r only
PVAR_FILE_PATH = INTERNAL_FILES_PATH / "fixed_target/pvar_files"
CS_FILES_PATH = INTERNAL_FILES_PATH / "fixed_target/cs"



"""
Zebra setup plans for extruder and fastchip serial collections.

For clarification on the Zebra setup in either use case, please see
https://confluence.diamond.ac.uk/display/MXTech/Zebra+settings+I24

Note on soft inputs. In the code, soft inputs are 1 indexed following the numbering on
the edm screen, while on the schematics they are 0 indexed. Thus, `Soft In 1` from the
schematics corresponds to soft_in_2 in the code.
"""

import bluesky.plan_stubs as bps
from dodal.devices.zebra.zebra import (
    ArmDemand,
    ArmSource,
    I24Axes,
    RotationDirection,
    SoftInState,
    TrigSource,
    Zebra,
)

from mx_bluesky.beamlines.i24.serial.log import SSX_LOGGER

# Detector specific outs
TTL_EIGER = 1
TTL_PILATUS = 2
TTL_FAST_SHUTTER = 4

SHUTTER_MODE = {
    "manual": SoftInState.NO,
    "auto": SoftInState.YES,
}

GATE_START = 1.0
SHUTTER_OPEN_TIME = 0.05  # For pp with long delays


def get_zebra_settings_for_extruder(
    exp_time: float,
    pump_exp: float,
    pump_delay: float,
) -> tuple[float, float]:
    """Calculates and returns gate width and step for extruder collections with pump \
    probe.

    The gate width is calculated by adding the exposure time, pump exposure and \
    pump delay. From this value, the gate step is obtained by adding a 0.01 buffer to \
    the width. The value of this buffer is empirically determined.
    """
    pump_probe_buffer = 0.01
    gate_width = pump_exp + pump_delay + exp_time
    gate_step = gate_width + pump_probe_buffer
    return gate_width, gate_step


def arm_zebra(zebra: Zebra):
    yield from bps.abs_set(zebra.pc.arm, ArmDemand.ARM, wait=True)
    SSX_LOGGER.info("Zebra armed.")


def disarm_zebra(zebra: Zebra):
    yield from bps.abs_set(zebra.pc.arm, ArmDemand.DISARM, wait=True)
    SSX_LOGGER.info("Zebra disarmed.")


def open_fast_shutter(zebra: Zebra):
    yield from bps.abs_set(zebra.inputs.soft_in_2, SoftInState.YES, wait=True)
    SSX_LOGGER.info("Fast shutter open.")


def close_fast_shutter(zebra: Zebra):
    yield from bps.abs_set(zebra.inputs.soft_in_2, SoftInState.NO, wait=True)
    SSX_LOGGER.info("Fast shutter closed.")


def set_shutter_mode(zebra: Zebra, mode: str):
    # SOFT_IN:B0 has to be disabled for manual mode
    yield from bps.abs_set(zebra.inputs.soft_in_1, SHUTTER_MODE[mode], wait=True)
    SSX_LOGGER.info(f"Shutter mode set to {mode}.")


def setup_pc_sources(
    zebra: Zebra,
    gate_source: TrigSource,
    pulse_source: TrigSource,
    group: str = "pc_sources",
):
    yield from bps.abs_set(zebra.pc.gate_source, gate_source, group=group)
    yield from bps.abs_set(zebra.pc.pulse_source, pulse_source, group=group)
    yield from bps.wait(group)


def setup_zebra_for_quickshot_plan(
    zebra: Zebra,
    exp_time: float,
    num_images: int,
    group: str = "setup_zebra_for_quickshot",
    wait: bool = True,
):
    """Set up the zebra for a static extruder experiment.

    Gate source set to 'External' and Pulse source set to 'Time'.
    The gate start is set to 1.0 and the gate width is calculated from \
    exposure time*number of images plus a 0.5 buffer.

    Args:
        zebra (Zebra): The zebra ophyd device.
        exp_time (float): Collection exposure time, in s.
        num_images (float): Number of images to be collected.
    """
    SSX_LOGGER.info("Setup ZEBRA for quickshot collection.")
    yield from bps.abs_set(zebra.pc.arm_source, ArmSource.SOFT, group=group)
    yield from setup_pc_sources(zebra, TrigSource.TIME, TrigSource.EXTERNAL)

    gate_width = exp_time * num_images + 0.5
    SSX_LOGGER.info(f"Gate start set to {GATE_START}, with width {gate_width}.")
    yield from bps.abs_set(zebra.pc.gate_start, GATE_START, group=group)
    yield from bps.abs_set(zebra.pc.gate_width, gate_width, group=group)

    yield from bps.abs_set(
        zebra.pc.gate_input, zebra.mapping.sources.SOFT_IN2, group=group
    )
    yield from bps.sleep(0.1)

    if wait:
        yield from bps.wait(group)
    SSX_LOGGER.info("Finished setting up zebra.")


def set_logic_gates_for_porto_triggering(
    zebra: Zebra, group: str = "porto_logic_gates"
):
    # To OUT2_TTL
    yield from bps.abs_set(
        zebra.logic_gates.and_gates[3].sources[1],
        zebra.mapping.sources.SOFT_IN2,
        group=group,
    )
    yield from bps.abs_set(
        zebra.logic_gates.and_gates[3].sources[2],
        zebra.mapping.sources.PULSE1,
        group=group,
    )
    # To OUT1_TTL
    yield from bps.abs_set(
        zebra.logic_gates.and_gates[4].sources[1],
        zebra.mapping.sources.SOFT_IN2,
        group=group,
    )
    yield from bps.abs_set(
        zebra.logic_gates.and_gates[4].sources[2],
        zebra.mapping.sources.PULSE2,
        group=group,
    )
    yield from bps.wait(group=group)


def setup_zebra_for_extruder_with_pump_probe_plan(
    zebra: Zebra,
    det_type: str,
    exp_time: float,
    num_images: int,
    pump_exp: float | None,
    pump_delay: float | None,
    pulse1_delay: float = 0.0,
    group: str = "setup_zebra_for_extruder_pp",
    wait: bool = True,
):
    """Zebra setup for extruder pump probe experiment using PORTO laser triggering.

    For this use case, both the laser and detector set up is taken care of by the Zebra.
    WARNING. This means that some hardware changes have been made.
    Because all four of the zebra ttl outputs are in use in this mode, when the \
    detector in use is the Eiger, the Pilatus cable is repurposed to trigger the light \
    source, and viceversa.

    The data collection output is OUT1_TTL for Eiger and OUT2_TTL for Pilatus and \
    should be set to AND3.

    Position compare settings:
        - The gate input is on SOFT_IN2.
        - The number of gates should be equal to the number of images to collect.
        - Gate source set to 'Time' and Pulse source set to 'External'.

    Pulse output settings:
        - Pulse1 is the laser control on the Zebra. It is set with a 0.0 delay and a \
            width equal to the requested laser dwell.
        - Pulse2 is the detector control. It is set with a delay equal to the laser \
            delay and a width equal to the exposure time.

    Args:
        zebra (Zebra): The zebra ophyd device.
        det_type (str): Detector in use, current choices are Eiger or Pilatus.
        exp_time (float): Collection exposure time, in s.
        num_images (int): Number of images to be collected.
        pump_exp (float): Laser dwell, in s.
        pump_delay (float): Laser delay, in s.
        pulse1_delay (float, optional): Delay to start pulse1 (the laser control) after \
            gate start. Defaults to 0.0.
    """
    SSX_LOGGER.info("Setup ZEBRA for pump probe extruder collection.")

    yield from set_shutter_mode(zebra, "manual")

    # Set gate to "Time" and pulse source to "External"
    yield from setup_pc_sources(zebra, TrigSource.TIME, TrigSource.EXTERNAL)

    # Logic gates
    yield from set_logic_gates_for_porto_triggering(zebra)

    # Set TTL out depending on detector type
    DET_TTL = TTL_EIGER if det_type == "eiger" else TTL_PILATUS
    LASER_TTL = TTL_PILATUS if det_type == "eiger" else TTL_EIGER
    yield from bps.abs_set(
        zebra.output.out_pvs[DET_TTL], zebra.mapping.sources.AND4, group=group
    )
    yield from bps.abs_set(
        zebra.output.out_pvs[LASER_TTL], zebra.mapping.sources.AND3, group=group
    )

    yield from bps.abs_set(
        zebra.pc.gate_input, zebra.mapping.sources.SOFT_IN2, group=group
    )

    assert pump_exp and pump_delay, "Must supply pump_exp and pump_delay!"
    gate_width, gate_step = get_zebra_settings_for_extruder(
        exp_time, pump_exp, pump_delay
    )
    SSX_LOGGER.info(
        f"""
        Gate start set to {GATE_START}, with calculated width {gate_width}
        and step {gate_step}.
        """
    )
    yield from bps.abs_set(zebra.pc.gate_start, GATE_START, group=group)
    yield from bps.abs_set(zebra.pc.gate_width, gate_width, group=group)
    yield from bps.abs_set(zebra.pc.gate_step, gate_step, group=group)
    # Number of gates is the same as the number of images
    yield from bps.abs_set(zebra.pc.num_gates, num_images, group=group)

    # Settings for extruder pump probe:
    # PULSE1_DLY is the start (0 usually), PULSE1_WID is the laser dwell set on edm
    # PULSE2_DLY is the laser delay set on edm, PULSE2_WID is the exposure time
    SSX_LOGGER.info(
        f"Pulse1 starting at {pulse1_delay} with width set to laser dwell {pump_exp}."
    )
    yield from bps.abs_set(
        zebra.output.pulse_1.input, zebra.mapping.sources.PC_GATE, group=group
    )
    yield from bps.abs_set(zebra.output.pulse_1.delay, pulse1_delay, group=group)
    yield from bps.abs_set(zebra.output.pulse_1.width, pump_exp, group=group)
    SSX_LOGGER.info(
        f"""
        Pulse2 starting at laser delay {pump_delay} with width set to \
        exposure time {exp_time}.
        """
    )
    yield from bps.abs_set(
        zebra.output.pulse_2.input, zebra.mapping.sources.PC_GATE, group=group
    )
    yield from bps.abs_set(zebra.output.pulse_2.delay, pump_delay, group=group)
    yield from bps.abs_set(zebra.output.pulse_2.width, exp_time, group=group)

    if wait:
        yield from bps.wait(group)
    SSX_LOGGER.info("Finished setting up zebra.")


def setup_zebra_for_fastchip_plan(
    zebra: Zebra,
    det_type: str,
    num_gates: int,
    num_exposures: int,
    exposure_time_s: float,
    start_time_offset: float = 0.0,
    group: str = "setup_zebra_for_fastchip",
    wait: bool = True,
):
    """Zebra setup for fixed-target triggering.

    For this use case, the laser set up is taken care of by the geobrick, leaving only \
    the detector side set up to the Zebra.
    The data collection output is OUT1_TTL for Eiger and OUT2_TTL for Pilatus and \
    should be set to AND3.

    Position compare settings:
        - The gate input is on IN3_TTL.
        - The number of gates should be equal to the number of apertures to collect.
        - Gate source set to 'External' and Pulse source set to 'Time'
        - Trigger source set to the exposure time with a 100us buffer in order to \
            avoid missing any triggers.
        - The trigger width is calculated depending on which detector is in use: the \
            Pilatus only needs the trigger rising edge to collect for a set time, while \
            the Eiger (used here in Externally Interrupter Exposure Series mode) \
            will only collect while the signal is high and will stop once a falling \
            edge is detected. For this reason a square wave pulse width will be set to \
            half the exposure time in the Pilatus case, and to the exposure time minus \
            a small drop (~100um) for the Eiger.

    Args:
        zebra (Zebra): The zebra ophyd device.
        det_type (str): Detector in use, current choices are Eiger or Pilatus.
        num_gates (int): Number of apertures to visit in a chip.
        num_exposures (int): Number of times data is collected in each aperture.
        exposure_time_s (float): Exposure time for each shot.
        start_time_offset (float): Delay on the start of the position compare. \
            Defaults to 0.0 (standard chip collection).
    """
    SSX_LOGGER.info("Setup ZEBRA for a fixed target collection.")

    yield from set_shutter_mode(zebra, "manual")

    yield from setup_pc_sources(zebra, TrigSource.EXTERNAL, TrigSource.TIME)

    # Logic Gates
    yield from bps.abs_set(
        zebra.logic_gates.and_gates[3].sources[1],
        zebra.mapping.sources.SOFT_IN2,
        group=group,
    )
    yield from bps.abs_set(
        zebra.logic_gates.and_gates[3].sources[2],
        zebra.mapping.sources.PC_PULSE,
        group=group,
    )

    yield from bps.abs_set(
        zebra.pc.gate_input, zebra.mapping.sources.IN3_TTL, group=group
    )

    # Set TTL out depending on detector type
    # And calculate some of the other settings
    if det_type == "eiger":
        yield from bps.abs_set(
            zebra.output.out_pvs[TTL_EIGER], zebra.mapping.sources.AND3, group=group
        )
    if det_type == "pilatus":
        yield from bps.abs_set(
            zebra.output.out_pvs[TTL_PILATUS], zebra.mapping.sources.AND3, group=group
        )

    # Square wave - needs a small drop to make it work for eiger
    pulse_width = (
        exposure_time_s - 0.0001 if det_type == "eiger" else exposure_time_s / 2
    )

    # 100us buffer needed to avoid missing some of the triggers
    exptime_buffer = exposure_time_s + 0.0001

    # Number of gates is the number of windows collected
    yield from bps.abs_set(zebra.pc.num_gates, num_gates, group=group)

    yield from bps.abs_set(zebra.pc.pulse_start, start_time_offset, group=group)
    yield from bps.abs_set(zebra.pc.pulse_step, exptime_buffer, group=group)
    yield from bps.abs_set(zebra.pc.pulse_width, pulse_width, group=group)
    yield from bps.abs_set(zebra.pc.pulse_max, num_exposures, group=group)

    if wait:
        yield from bps.wait(group)
    SSX_LOGGER.info("Finished setting up zebra.")


def open_fast_shutter_at_each_position_plan(
    zebra: Zebra,
    num_exposures: int,
    exposure_time_s: float,
    group: str = "fast_shutter_control",
    wait: bool = True,
):
    """A plan to control the fast shutter so that it will open at each position.
    This plan is a specific setup for pump probe fixed target triggering with long \
    delays between exposures.

    For this use case, the fast shutter opens and closes at every position to avoid \
    destroying the crystals by exposing them to the beam for a long time in between \
    collections.

    The shutter opening time, hardcoded to 0.05, has been empirically determined.

    Fast shutter (Pulse2) output settings:
        - Output is OUT4_TTL set to PULSE2.
        - Pulse2 is set with a delay equal to 0 and a width equal to the exposure time \
            multiplied by the number of exposures, plus the shutter opening time.

    Args:
        zebra (Zebra): The zebra ophyd device.
        num_exposures (int): Number of times data is collected in each aperture.
        exposure_time_s (float): Exposure time for each shot.
    """
    SSX_LOGGER.info(
        "ZEBRA setup for fastchip collection with long delays between exposures."
    )
    SSX_LOGGER.debug("Controlling the fast shutter on PULSE2.")
    # Output panel pulse_2 settings
    yield from bps.abs_set(
        zebra.output.pulse_2.input, zebra.mapping.sources.PC_GATE, group=group
    )
    yield from bps.abs_set(zebra.output.pulse_2.delay, 0.0, group=group)
    pulse2_width = num_exposures * exposure_time_s + SHUTTER_OPEN_TIME
    yield from bps.abs_set(zebra.output.pulse_2.width, pulse2_width, group=group)

    # Fast shutter
    yield from bps.abs_set(
        zebra.output.out_pvs[TTL_FAST_SHUTTER],
        zebra.mapping.sources.PULSE2,
        group=group,
    )

    if wait:
        yield from bps.wait(group=group)
    SSX_LOGGER.debug("Finished setting up for long delays.")


def reset_pc_gate_and_pulse(zebra: Zebra, group: str = "reset_pc"):
    yield from bps.abs_set(zebra.pc.gate_start, 0, group=group)
    yield from bps.abs_set(zebra.pc.pulse_width, 0, group=group)
    yield from bps.abs_set(zebra.pc.pulse_step, 0, group=group)
    yield from bps.wait(group=group)


def reset_output_panel(zebra: Zebra, group: str = "reset_zebra_outputs"):
    # Reset TTL out
    yield from bps.abs_set(
        zebra.output.out_pvs[2], zebra.mapping.sources.PC_GATE, group=group
    )
    yield from bps.abs_set(
        zebra.output.out_pvs[3], zebra.mapping.sources.DISCONNECT, group=group
    )
    yield from bps.abs_set(
        zebra.output.out_pvs[4], zebra.mapping.sources.OR1, group=group
    )

    yield from bps.abs_set(
        zebra.output.pulse_1.input, zebra.mapping.sources.DISCONNECT, group=group
    )
    yield from bps.abs_set(
        zebra.output.pulse_2.input, zebra.mapping.sources.DISCONNECT, group=group
    )

    yield from bps.wait(group=group)


def zebra_return_to_normal_plan(
    zebra: Zebra, group: str = "zebra-return-to-normal", wait: bool = True
):
    """A plan to reset the Zebra settings at the end of a collection.

    This plan should only be run after disarming the Zebra.
    """
    yield from bps.abs_set(zebra.pc.reset, 1, group=group)

    # Reset PC_GATE and PC_SOURCE to "Position"
    yield from setup_pc_sources(zebra, TrigSource.POSITION, TrigSource.POSITION)

    yield from bps.abs_set(
        zebra.pc.gate_input, zebra.mapping.sources.SOFT_IN3, group=group
    )
    yield from bps.abs_set(zebra.pc.num_gates, 1, group=group)
    yield from bps.abs_set(
        zebra.pc.pulse_input, zebra.mapping.sources.DISCONNECT, group=group
    )

    # Logic Gates
    yield from bps.abs_set(
        zebra.logic_gates.and_gates[3].sources[1],
        zebra.mapping.sources.PC_ARM,
        group=group,
    )
    yield from bps.abs_set(
        zebra.logic_gates.and_gates[3].sources[2],
        zebra.mapping.sources.IN1_TTL,
        group=group,
    )

    # Reset TTL out
    yield from reset_output_panel(zebra)

    # Reset Pos Trigger and direction to rotation axis ("omega") and positive
    yield from bps.abs_set(zebra.pc.gate_trigger, I24Axes.OMEGA.value, group=group)
    yield from bps.abs_set(zebra.pc.dir, RotationDirection.POSITIVE, group=group)

    #
    yield from reset_pc_gate_and_pulse(zebra)

    if wait:
        yield from bps.wait(group)
    SSX_LOGGER.info("Zebra settings back to normal.")


def reset_zebra_when_collection_done_plan(zebra: Zebra):
    """
    End of collection zebra operations: close fast shutter, disarm and reset settings.
    """
    SSX_LOGGER.debug("Close the fast shutter.")
    yield from close_fast_shutter(zebra)
    SSX_LOGGER.debug("Disarm the zebra.")
    yield from disarm_zebra(zebra)
    SSX_LOGGER.debug("Set zebra back to normal.")
    yield from zebra_return_to_normal_plan(zebra, wait=True)
